# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that force field files are properly read.
"""

import textwrap
import pytest
import numpy as np
import vermouth.ffinput
import vermouth.forcefield
import vermouth.molecule
import vermouth.gmx.itp_read

class TestBlock:
    @staticmethod
    def test_moleculetype():
        lines = """
        [ moleculetype ]
        GLY  3
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        assert 'GLY' in ff.blocks
        assert ff.blocks['GLY'].nrexcl == 3

    @staticmethod
    def test_multiple_moleculetype():
        lines = """
            [ moleculetype ]
            GLY  3

            [ moleculetype ]
            VAL 2
            """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        assert set(ff.blocks) == {'GLY', 'VAL'}
        assert ff.blocks['GLY'].nrexcl == 3
        assert ff.blocks['VAL'].nrexcl == 2

    @staticmethod
    def test_atoms():
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
          1 P4 1 ALA BB 1
        2 P3\t1\tALA   SC1 2 -3 7
        3 P2 1 ALA SC2 3 {"custom": 4, "other": "plop"}
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        block = ff.blocks['GLY']
        assert len(block.nodes) == 3
        assert block.nodes['BB'] == {'atomname': 'BB', 'atype': 'P4',
                                     'resname': 'ALA', 'resid': 1,
                                     'charge_group': 1}
        assert block.nodes['SC1'] == {'atomname': 'SC1', 'atype': 'P3',
                                      'resname': 'ALA', 'resid': 1,
                                      'charge_group': 2, 'charge': -3.0,
                                      'mass': 7.0}
        assert block.nodes['SC2'] == {'atomname': 'SC2', 'atype': 'P2',
                                      'resname': 'ALA', 'resid': 1,
                                      'charge_group': 3, 'custom': 4,
                                      'other': 'plop'}

    @staticmethod
    def test_fixed_number_interaction():
        """
        Define an interaction for which the number of atoms required is known.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P3 1 ALA SC1 2
        3 P2 1 ALA SC2 3

        [ bonds ]
        1 2 1 0.2 100
        2 3 4 0.6 700 {"comment": "custom"}
        3 1 -- 9 plop toto
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        block = ff.blocks['GLY']

        bonds = [
            vermouth.molecule.Interaction(
                atoms=['BB', 'SC1'], parameters=['1', '0.2', '100'], meta={},
            ),
            vermouth.molecule.Interaction(
                atoms=['SC1', 'SC2'], parameters=['4', '0.6', '700'],
                meta={'comment': 'custom'},
            ),
            vermouth.molecule.Interaction(
                atoms=['SC2', 'BB'], parameters=['9', 'plop', 'toto'], meta={},
            ),
        ]
        assert block.interactions['bonds'] == bonds


    @staticmethod
   #def test_variable_number_interaction():
   #    """
   #    Define an interaction for which the number of atoms required is unknown.
   #    """
   #    lines = """
   #    [ moleculetype ]
   #    GLY 1

   #    [ atoms ]
   #    1 P4 1 ALA BB 1
   #    2 P3 1 ALA SC1 2
   #    3 P2 1 ALA SC2 3

   #    [ unknown ]
   #    1 2 -- 1 0.2 100
   #    2 3 1 -- 0.6 700 {"comment": "custom"}
   #    3 1 2
   #    """
   #    lines = textwrap.dedent(lines).splitlines()
   #    ff = vermouth.forcefield.ForceField(name='test_ff')
   #    vermouth.ffinput.read_ff(lines, ff)
   #    block = ff.blocks['GLY']

   #    interactions = [
   #        vermouth.molecule.Interaction(
   #            atoms=['BB', 'SC1'], parameters=['1', '0.2', '100'], meta={},
   #        ),
   #        vermouth.molecule.Interaction(
   #            atoms=['SC1', 'SC2', 'BB'], parameters=['0.6', '700'],
   #            meta={'comment': 'custom'},
   #        ),
   #        vermouth.molecule.Interaction(
   #            atoms=['SC2', 'BB', 'SC1'], parameters=[], meta={},
   #        ),
   #    ]
   #    assert block.interactions['unknown'] == interactions

    @staticmethod
    def test_interaction_by_name():
        """
        Create an interaction and refer to atoms by name.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P3 1 ALA SC1 2
        3 P2 1 ALA SC2 3

        [ bonds ]
        BB SC1 1 0.2 100
        SC1 SC2 4 0.6 700 {"comment": "custom"}
        SC2 BB -- 9 plop toto
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        block = ff.blocks['GLY']

        bonds = [
            vermouth.molecule.Interaction(
                atoms=['BB', 'SC1'], parameters=['1', '0.2', '100'], meta={},
            ),
            vermouth.molecule.Interaction(
                atoms=['SC1', 'SC2'], parameters=['4', '0.6', '700'],
                meta={'comment': 'custom'},
            ),
            vermouth.molecule.Interaction(
                atoms=['SC2', 'BB'], parameters=['9', 'plop', 'toto'], meta={},
            ),
        ]
        assert block.interactions['bonds'] == bonds

    # test_interaction_edges(interaction_lines, edges)
    @staticmethod
    @pytest.mark.parametrize(
        'interaction_lines, edges', (
            (  # regular bonds create edges
                """
                [ bonds ]
                BB SC1
                SC2 SC3
                """,
                (('BB', 'SC1'), ('SC2', 'SC3')),
            ),
            (  # regular angles create edges
                """
                [ angles ]
                BB SC1 SC2
                SC1 SC2 SC3
                """,
                (('BB', 'SC1'), ('SC1', 'SC2'), ('SC2', 'SC3')),
            ),
            (  # regular dihedrals create edges
                """
                [dihedrals]
                BB SC1 SC2 SC3
                """,
                (('BB', 'SC1'), ('SC1', 'SC2'), ('SC2', 'SC3')),
            ),
            (  # impropers do not create edges
                """
                [ impropers ]
                BB SC1 SC2 SC3
                """,
                (),
            ),
            (
                """
                [ dihedrals ]
                BB SC1 SC2 SC3 2 ; gromacs dihedrals of type 2 are impropers
                """,
                (),
            ),
            (  # the edge section creates edges
                """
                [ edges ]
                SC1 SC2
                BB SC3
                """,
                (('SC1', 'SC2'), ('BB', 'SC3')),
            ),
            (  # edges can be deactivated from interactions
                """
                [ bonds ]
                BB SC1 -- {"edge": false}
                SC2 SC3 -- {"edge": true}
                """,
                (('SC2', 'SC3'), ),
            ),
        )
    )
    def test_interaction_edges(interaction_lines, edges):
        """
        Edges are created where expected.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P3 1 ALA SC1 2
        3 P2 1 ALA SC2 3
        4 P2 1 ALA SC3 3
        """
        lines = textwrap.dedent(lines) + textwrap.dedent(interaction_lines)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        block = ff.blocks['GLY']

        assert (set(frozenset(edge) for edge in block.edges)
                == set(frozenset(edge) for edge in edges))

    # test_interaction_fail_reference
    @staticmethod
    @pytest.mark.parametrize('interaction_lines', (
        # Refers, by index, to an atom out of range
        """
        [ bonds ]
        1 6
        """,
        # Refers, by name, to a non-defined atom (SC2 does not exist)
        """
        [ bonds ]
        BB SC2
        """,
        # One missing atom in the definition (refers to only one atom instead of 2)
        """
        [ bonds ]
        BB
        """,
        # Prefixed atom in a block interaction
        """
        [ bonds ]
        BB +SC1
        """,
        """
        [ bonds ]
        BB -SC1
        """,
        """
        [ bonds ]
        BB <SC1
        """,
        """
        [ bonds ]
        BB >SC1
        """,
    ))
    def test_interaction_fail_reference(interaction_lines):
        """
        Faulty atom references in interactions lead to an exception.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        ; The following atoms have names that should not be valid
        2 P4 2 TRP +SC1 1
        3 P4 2 TRP -SC1 1
        4 P4 2 TRP <SC1 1
        5 P4 2 TRP >SC1 1
        """
        lines = textwrap.dedent(lines)
        lines = lines.splitlines()
        lines += textwrap.dedent(interaction_lines).splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        with pytest.raises(IOError):
            vermouth.ffinput.read_ff(lines, ff)

    @staticmethod
    def test_interaction_hash_meta():
        """
        Make sure that the `#meta` lines are accounted for.
        """
        lines = """
        [ moleculetype ]
        GLY  3

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1
        3 P4 1 ALA SC2 1
        4 P4 1 ALA SC3 1
        5 P4 1 ALA SC4 1

        [ bonds ]
        1 2 -- {"custom": 0}
        #meta {"first meta": "a", "first meta 2": "b"}
        2 3 -- {"custom": 1}
        3 4
        #meta {"first meta": 10, "second meta": 20}
        4 5
        """
        lines = textwrap.dedent(lines)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        interactions = ff.blocks['GLY'].interactions['bonds']
        meta = [interaction.meta for interaction in interactions]
        expected = [
            {'custom': 0},
            {'first meta': 'a', 'first meta 2': 'b', 'custom': 1},
            {'first meta': 'a', 'first meta 2': 'b'},
            {'first meta': 10, 'first meta 2': 'b', 'second meta': 20},
        ]
        assert meta == expected

    @staticmethod
    @pytest.mark.parametrize('meta_line', (
        # Extra column
        '#meta {"a": 0, "b": 1} "plop"',
        # Missing column
        '#meta',
        # The value is not a dict
        '#meta "plop"'
    ))
    def test_invalid_meta(meta_line):
        """
        #meta lines fail when misformatted.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1

        [ bonds ]
        {}
        BB SC1
        """
        lines = textwrap.dedent(lines).format(meta_line)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        with pytest.raises(IOError):
            vermouth.ffinput.read_ff(lines, ff)

    @staticmethod
    @pytest.mark.parametrize('section_name', (
        'non-edges', 'patterns', 'features', '!bonds', '!angles'
    ))
    def test_invalid_section_fails(section_name):
        """
        Using the sections that are invalid in a block fails.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1

        [ {} ]
        BB SC1
        """
        lines = textwrap.dedent(lines).format(section_name)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        with pytest.raises(IOError):
            vermouth.ffinput.read_ff(lines, ff)

    @staticmethod
    def test_atomname_unique():
        """
        Atom names are unique in a block.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA BB 1
        """
        lines = textwrap.dedent(lines)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        with pytest.raises(IOError):
            vermouth.ffinput.read_ff(lines, ff)

    @staticmethod
    @pytest.mark.parametrize('interaction_line, expected', (
        (
            "BB SC1 1 dist(BB,SC1)",
            ['1', vermouth.molecule.ParamDistance(['BB', 'SC1'])],
        ),
        (
            "BB SC1 2 dist(BB,SC1|.3f) 6",
            ['2', vermouth.molecule.ParamDistance(['BB', 'SC1'], format_spec='.3f'), '6'],
        ),
    ))
    def test_param_effector(interaction_line, expected):
        """
        Parameter effectors are detected and instanciated.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1

        [ bonds ]
        {}
        """
        lines = textwrap.dedent(lines.format(interaction_line))
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        parameters = ff.blocks['GLY'].interactions['bonds'][0].parameters

        assert parameters == expected

    @staticmethod
    def test_param_effector_fail():
        """
        A non existing parameter effector causes the parser to fail.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1

        [ bonds ]
        BB SC1 7 inexistant(BB,SC1)
        """
        lines = textwrap.dedent(lines)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        with pytest.raises(IOError):
            vermouth.ffinput.read_ff(lines, ff)

    @staticmethod
    @pytest.mark.parametrize('interaction_lines, expected', (
        [
            """
            [ dihedrals ]
            BB SC1 SC2 SC3 2 1 2 3 4
            BB SC1 SC2 SC4 1
            """,
            {'impropers': [vermouth.molecule.Interaction(atoms='BB SC1 SC2 SC3'.split(),
                                                         parameters='2 1 2 3 4'.split(),
                                                         meta={})],
             'dihedrals': [vermouth.molecule.Interaction(atoms='BB SC1 SC2 SC4'.split(),
                                                         parameters=['1'],
                                                         meta={})]}
        ],
        [
            """
            [ dihedrals ]
            BB SC1 SC2 SC3 1 1 2 3 4
            BB SC1 SC2 SC4 1
            """,
            {'dihedrals': [vermouth.molecule.Interaction(atoms='BB SC1 SC2 SC3'.split(),
                                                         parameters='1 1 2 3 4'.split(),
                                                         meta={}),
                           vermouth.molecule.Interaction(atoms='BB SC1 SC2 SC4'.split(),
                                                         parameters=['1'],
                                                         meta={})],
             'impropers': []}
        ],
        [
            """
            [ dihedrals ]
            BB SC1 SC2 SC3 2 1 2 3 4
            BB SC1 SC2 SC4 2
            """,
            {'impropers': [vermouth.molecule.Interaction(atoms='BB SC1 SC2 SC3'.split(),
                                                         parameters='2 1 2 3 4'.split(),
                                                         meta={}),
                           vermouth.molecule.Interaction(atoms='BB SC1 SC2 SC4'.split(),
                                                         parameters=['2'],
                                                         meta={})],
             'dihedrals': []}
        ],
    ))
    def test_multiple_impropers(interaction_lines, expected):
        lines = """
        [ moleculetype ]
        XOXO 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P3 1 ALA SC1 2
        3 P2 1 ALA SC2 3
        4 P2 1 ALA SC3 3
        5 P2 1 ALA SC4 3
        """
        lines = textwrap.dedent(lines) + textwrap.dedent(interaction_lines)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        assert ff.blocks['XOXO'].interactions == expected


class TestLink:
    @staticmethod
    def test_link():
        """
        The minimum link works.
        """
        lines = """
        [ link ]

        [link]
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        assert len(ff.links) == 2

    @staticmethod
    @pytest.mark.parametrize('lines, atoms, interactions, edges', (
        (  # Base case
            """
            [link]
            [bonds]
            BB SC1 1 0.3 100
            """,
            (
                ('BB', {'atomname': 'BB', 'order': 0}),
                ('SC1', {'atomname': 'SC1', 'order': 0}),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['BB', 'SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={},
                ),
            ]},
            (('BB', 'SC1'),),
        ),
        (  # Base case with prefix
            """
            [link]
            [bonds]
            BB +SC1 1 0.3 100
            """,
            (
                ('BB', {'atomname': 'BB', 'order': 0}),
                ('+SC1', {'atomname': 'SC1', 'order': 1}),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['BB', '+SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={},
                ),
            ]},
            (('BB', '+SC1'),),
        ),
        (  # Base case with prefix
            """
            [link]
            [bonds]
            ---BB SC1 1 0.3 100
            """,
            (
                ('---BB', {'atomname': 'BB', 'order': -3}),
                ('SC1', {'atomname': 'SC1', 'order': 0}),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['---BB', 'SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={},
                ),
            ]},
            (('---BB', 'SC1'),),
        ),
        (  # Base case with prefix
            """
            [link]
            [bonds]
            <<BB >SC1 1 0.3 100
            """,
            (
                ('<<BB', {'atomname': 'BB', 'order': '<<'}),
                ('>SC1', {'atomname': 'SC1', 'order': '>'}),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['<<BB', '>SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={},
                ),
            ]},
            (('<<BB', '>SC1'),),
        ),
        (  # Base case with implicit order in the attributes
            """
            [link]
            [bonds]
            BB SC1 {"order": 1} 1 0.3 100
            """,
            (
                ('BB', {'atomname': 'BB', 'order': 0}),
                ('+SC1', {'atomname': 'SC1', 'order': 1}),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['BB', '+SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={},
                ),
            ]},
            (('BB', '+SC1'),),
        ),
        (  # Base case with implicit order in the attributes
            """
            [link]
            [bonds]
            BB SC1 {"order": -2} 1 0.3 100
            """,
            (
                ('BB', {'atomname': 'BB', 'order': 0}),
                ('--SC1', {'atomname': 'SC1', 'order': -2}),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['BB', '--SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={},
                ),
            ]},
            (('BB', '--SC1'),),
        ),
        (  # Base case with implicit order in the attributes
            """
            [link]
            [bonds]
            SC1 SC1 {"order": ">>"} 1 0.3 100
            """,
            (
                ('SC1', {'atomname': 'SC1', 'order': 0}),
                ('>>SC1', {'atomname': 'SC1', 'order': '>>'}),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['SC1', '>>SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={},
                ),
            ]},
            (('SC1', '>>SC1'),),
        ),
        (  # Inhibit edges
            """
            [link]
            [bonds]
            BB SC1 1 0.3 100 {"edge": false}
            """,
            (
                ('BB', {'atomname': 'BB', 'order': 0}),
                ('SC1', {'atomname': 'SC1', 'order': 0}),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['BB', 'SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={'edge': False},
                ),
            ]},
            (),
        ),
        (  # Atom already in the link
            """
            [link]
            [atoms]
            BB {"element": "Q"}
            [bonds]
            BB SC1 1 0.3 100
            """,
            (
                ('BB', {'atomname': 'BB', 'order': 0, 'element': 'Q'}),
                ('SC1', {'atomname': 'SC1', 'order': 0}),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['BB', 'SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={},
                ),
            ]},
            (('BB', 'SC1'),),
        ),
        (  # Interaction atoms have attributes
            """
            [link]
            [bonds]
            BB {"custom": 1, "plop": "toto"} SC1 {"toto": "tata"} 1 0.3 100
            """,
            (
                ('BB', {'atomname': 'BB', 'order': 0, 'custom': 1, 'plop': 'toto'}),
                ('SC1', {'atomname': 'SC1', 'order': 0, 'toto': 'tata'}),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['BB', 'SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={},
                ),
            ]},
            (('BB', 'SC1'),),
        ),
        (  # Atom attributes use a Choice
            """
            [link]
            [bonds]
            BB SC1 {"resname": "GLY|ALA|VAL"} 1 0.3 100
            """,
            (
                ('BB', {'atomname': 'BB', 'order': 0}),
                ('SC1', {
                    'atomname': 'SC1', 'order': 0,
                    'resname': vermouth.molecule.Choice(['GLY', 'ALA', 'VAL']),
                }),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['BB', 'SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={},
                ),
            ]},
            (('BB', 'SC1'),),
        ),
        (  # Atom attributes use a LinkPredicate defined as link attribute
            """
            [link]
            resname not("VAL")
            [bonds]
            BB SC1 1 0.3 100
            """,
            (
                ('BB', {
                    'atomname': 'BB', 'order': 0,
                    'resname': vermouth.molecule.NotDefinedOrNot('VAL'),
                }),
                ('SC1', {
                    'atomname': 'SC1', 'order': 0,
                    'resname': vermouth.molecule.NotDefinedOrNot('VAL'),
                }),
            ),
            {'bonds': [
                vermouth.molecule.Interaction(
                    atoms=['BB', 'SC1'],
                    parameters=['1', '0.3', '100'],
                    meta={},
                ),
            ]},
            (('BB', 'SC1'),),
        ),
    ))
    def test_link_interactions(lines, atoms, interactions, edges):
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        link = ff.links[0]
        assert tuple(link.nodes(data=True)) == atoms
        assert link.interactions == interactions
        assert (set(frozenset(edge) for edge in link.edges)
                == set(frozenset(edge) for edge in edges))


    @staticmethod
    def test_number_of_links():
        """
        Links are stored in a list so unlike blocks
        they are not overwritten when they are returned
        from the parsers. Thus we should check not only
        that the expected links are in but also how many
        there are.
        """
        lines = """
        [ moleculetype ]
        GLY 1
        [ atoms ]
        1 P4 1 ALA BB 1
        [ link ]
        [ bonds ]
        BB +BB params
        BB SC1 params
        [ angles ]
        BB +BB ++BB params
        [ link ]
        [ bonds ]
        BB SC2 params
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)

        assert len(ff.links) == 2

    @staticmethod
    def test_negative_interactions():
        """
        Prefixing an interaction section name with a ! adds the interactions to
        the ones to delete.
        """
        lines = """
        [ link ]
        [ bonds ]
        BB SC2
        [ !bonds ]
        BB SC1 1 20 3 {"toto": "tata"}
        SC1 {"attr": 8} SC2
        ;[ !custom ]
        ;BB SC1 {"attr_plop": 9} SC2 -- {"plop": 0, "foo": "bar"}
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        link = ff.links[0]
        assert link.interactions == {
            "bonds": [
                vermouth.molecule.Interaction(
                    atoms=['BB', 'SC2'],
                    parameters=[],
                    meta={},
                ),
            ],
        }
        assert link.removed_interactions == {
            "bonds": [
                vermouth.molecule.DeleteInteraction(
                    atoms=['BB', 'SC1'],
                    atom_attrs=[{}, {}],
                    parameters=['1', '20', '3'],
                    meta={'toto': 'tata'},
                ),
                vermouth.molecule.DeleteInteraction(
                    atoms=['SC1', 'SC2'], parameters=[], meta={},
                    atom_attrs=[{'attr': 8}, {}],
                ),
            ],
            #"custom": [
            #    vermouth.molecule.DeleteInteraction(
            #        atoms=['BB', 'SC1', 'SC2'],
            #        atom_attrs=[{}, {'attr_plop': 9}, {}],
            #        parameters=[],
            #        meta={'plop': 0, 'foo': 'bar'},
            #    ),
            #],
        }

    @staticmethod
    def test_features():
        """
        Links accept features.
        """
        lines = """
        [ link ]
        [ features ]
        toto tata
        plop
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        link = ff.links[0]
        assert link.features == {'toto', 'tata', 'plop'}

    @staticmethod
    def test_patterns():
        """
        Links accept patterns.
        """
        lines = """
        [ link ]
        [ patterns ]
        BB {"custom": 0} SC1 SC2 {"plop": "toto"}
        BB SC1 {"custom": "A|BB|CCC"} SC2
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        link = ff.links[0]
        assert link.patterns == [
            [['BB', {'custom': 0}], ['SC1', {}], ['SC2', {'plop': 'toto'}],],
            [
                ['BB', {}],
                ['SC1', {'custom': vermouth.molecule.Choice(['A', 'BB', 'CCC'])}],
                ['SC2', {}],
            ],
        ]

    @staticmethod
    def test_attributes():
        """
        Link attributes are added to the atoms.
        """
        lines = """
        [ link ]
        attr0 "plop"
        attr1 {"foo": "bar"}
        attr2 10
        attr3 "A|BB|CCC"
        attr4 not("stuff")
        [ atoms ]
        XXX {"attr0": "other"}  ; overwrite attr0
        [ bonds ]
        YYY ZZZ
        QQQ {"attr1": 33} WWW  ; overwrite attr1
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        link = ff.links[0]
        # Check the overwritten attributes first
        assert link.nodes['XXX']['attr0'] == 'other'
        assert link.nodes['QQQ']['attr1'] == 33
        assert all(link.nodes[key]['attr0'] == 'plop'
                   for key in ('YYY', 'ZZZ', 'QQQ', 'WWW'))
        assert all(link.nodes[key]['attr1'] == {'foo': 'bar'}
                   for key in ('YYY', 'ZZZ', 'XXX', 'WWW'))
        # Check the others
        assert all(link.nodes[key]['attr2'] == 10 for key in link.nodes)
        assert all(
            link.nodes[key]['attr3'] == vermouth.molecule.Choice(['A', 'BB', 'CCC'])
            for key in link.nodes
        )
        assert all(
            link.nodes[key]['attr4'] == vermouth.molecule.NotDefinedOrNot('stuff')
            for key in link.nodes
        )

    @staticmethod
    def test_moleta():
        """
        The [ molmeta ] section is read.
        """
        lines = """
        [ link ]
        [ molmeta ]
        attr0 10
        attr1 "plop"
        attr2 "A|BB|CCC"
        attr3 not("toto")
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        link = ff.links[0]
        assert link.molecule_meta == {
            'attr0': 10,
            'attr1': 'plop',
            'attr2': vermouth.molecule.Choice(['A', 'BB', 'CCC']),
            'attr3': vermouth.molecule.NotDefinedOrNot('toto'),
        }

    @staticmethod
    def test_non_edges():
        lines = """
        [ link ]
        [ non-edges ]
        XXX YYY {"atomname": "notYYY"}
        ABC +DEF
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        link = ff.links[0]
        assert link.non_edges == [
            ['XXX', {'atomname': 'notYYY', 'order': 0}],
            ['ABC', {'atomname': 'DEF', 'order': 1}],
        ]



class TestModification:
    @staticmethod
    def test_modification():
        """
        The modification section creates a modification.
        """
        lines = """
        [ modification ]
        [modification]
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        # 2 modifications were made, but they both had the same name. This
        # should probably raise an error somewhere.
        assert list(ff.modifications.keys()) == ['']

    @staticmethod
    def test_example():
        lines = """
        [ modification ]
        C-ter
        [ atoms ]
        CA{"element": "C"}  ; the space between the name and the attributes is optionnal
        C {"element": "C"}
        O {"element": "O"}
        OXT {"element": "O", "PTM_atom": true, "replace": {"atomname": null}}

        [ edges ]
        CA C
        C O
        C OXT
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.ffinput.read_ff(lines, ff)
        modification = ff.modifications['C-ter']

        assert modification.name == 'C-ter'
        assert tuple(modification.nodes(data=True)) == (
            ('CA', {'element': 'C', 'PTM_atom': False, 'atomname': 'CA'}),
            ('C', {'element': 'C', 'PTM_atom': False, 'atomname': 'C'}),
            ('O', {'element': 'O', 'PTM_atom': False, 'atomname': 'O'}),
            ('OXT', {
                'element': 'O',
                'PTM_atom': True,
                'atomname': 'OXT',
                'replace': {'atomname': None}
            }),
        )
        assert set(frozenset(edge) for edge in modification.edges) == {
            frozenset(('CA', 'C')),
            frozenset(('C', 'O')),
            frozenset(('C', 'OXT')),
        }


def test_variables():
    """
    Test that the `[ variables ]` section is read as expected.
    """
    lines = """
        [ variables ]
        integer 1
        float 0.3
        string "mass"
        dictionary {"a": 1, "b": {"plop": "toto"}}
    """
    lines = textwrap.dedent(lines).splitlines()
    ff = vermouth.forcefield.ForceField(name='test_ff')
    vermouth.ffinput.read_ff(lines, ff)

    # We test the float first as we want to test its equality with a tolerance.
    # Then we can test that the rest is equal.
    assert np.allclose(ff.variables['float'], 0.3)
    del ff.variables['float']
    assert ff.variables == {'integer': 1, 'string': 'mass',
                            'dictionary': {'a': 1, 'b': {'plop': 'toto'}}}
test_variables()

def test_variables_existing():
    """
    Test that the `[ variables ]` section updates existing variables.
    """
    lines = """
        [ variables ]
        added 1
        updated 2
    """
    lines = textwrap.dedent(lines).splitlines()
    ff = vermouth.forcefield.ForceField(name='test_ff')
    ff.variables['existing'] = 0
    ff.variables['updated'] = 5
    vermouth.ffinput.read_ff(lines, ff)
    assert ff.variables == {'existing': 0, 'added': 1, 'updated': 2}


def test_variables_macro():
    """
    Test that macros are used when defining variables.
    """
    lines = """
        [ macros ]
        int_macro 1
        quote_macro "quoted"
        unquoted_macro unquoted
        int_macro2 2

        [ variables ]
        int_variable $int_macro
        quote_variable $quote_macro
        unquoted_variable $unquoted_macro
        multiple $int_macro$int_macro2
    """
    lines = textwrap.dedent(lines).splitlines()
    ff = vermouth.forcefield.ForceField(name='test_ff')
    vermouth.ffinput.read_ff(lines, ff)
    assert ff.variables == {'int_variable': 1, 'quote_variable': 'quoted',
                            'unquoted_variable': 'unquoted', 'multiple': 12}

def test_variable_first():
    """
    The [ variables ] section must come before blocks, links, or modifications.
    """
    lines = """
        [ macros ]
        int_macro 1
        quote_macro "quoted"
        unquoted_macro unquoted

        [ moleculetype ]
        GLY 2

        [ variables ]
        int_variable $int_macro
        quote_variable $quote_macro
        unquoted_variable $unquoted_macro
    """
    lines = textwrap.dedent(lines).splitlines()
    ff = vermouth.forcefield.ForceField(name='test_ff')
    with pytest.raises(IOError):
        vermouth.ffinput.read_ff(lines, ff)


@pytest.mark.parametrize('lines', (
    # Misformated section title
    """
    [ macros
    """,
    """
    macros]
    """,
    """
    [; macros]
    """,
    # Misformatted attributes
    """
    [ variables ]
    toto {"unfinished": 0
    """,
    """
    [ variables ]
    toto {"unfinished": "plop": 5}}
    """,
    # Misformatted macros
    """
    [ macros ]
    toto tata plop ;  too many columns
    """,
    """
    [ macros ]
    toto  ; not enough columns
    """,
    # Misformatted variables
    """
    [ variables ]
    toto tata plop ;  too many columns
    """,
    """
    [ variables ]
    toto  ; not enough columns
    """,
    # Multiple -- separators
    """
    [ moleculetype ]
    GLY 1

    [ atoms ]
    1 P4 1 ALA BB 1
    2 P4 1 ALA SC1 1

    [ bonds ]
    BB SC1 -- 7 8 -- 9 10
    """,
    # Atom attributes without reference
    """
    [ link ]
    [ patterns ]
    {"atomname": "CA"}
    """,
    # Prefix is inconsistent
    """
    [ link ]
    [ bonds ]
    +-+XXX YYY
    """,
    """
    [ link ]
    [ bonds ]
    ><XXX YYY
    """,
    # Prefix is missing a reference
    """
    [ link ]
    [ bonds ]
    +++ YYY
    """,
    # Invalid order
    """
    [ link ]
    [ atoms ]
    XXX {"order": "invalid"}
    """,
    # Order is inconsistent with the prefix
    """
    [ link ]
    [ atoms ]
    ++XXX {"order": -3}
    """,
    # Mismatch between attributes in different sections
    """
    [ link ]
    [ atoms ]
    XXX {"custom": 0}
    [ bonds ]
    XXX {"custom": 1} YYY
    """,
    """
    [ link ]
    [ bonds ]
    XXX {"custom": 1} YYY
    XXX {"custom": 2} ZZZ
    """,
    # Wrong number of tokens in the link [ atoms ] section
    """
    [ link ]
    [ atoms ]
    XXX
    """,
    """
    [ link ]
    [ atoms ]
    XXX {"custom": 0} "extra"
    """,
    # Wrong number of tokens in the [ molmeta ] section
    """
    [ link ]
    [ molmeta ]
    XXX
    """,
    """
    [ link ]
    [ molmeta ]
    1 2 3
    """,
    # [ molmeta ] section outside a link
    """
    [ molmeta ]
    XXX 2
    """,
    """
    [ moleculetype ]
    ABC   3
    [ molmeta ]
    XXX 2
    """,
    # [ atoms ] section not in an authorized section
    """
    [ atoms ]
    XXX {"custom": 0}
    """,
    """
    [ variables ]
    XXX 10
    [ atoms ]
    ABD {"custom": 0}
    """,
    # Specify edge between unknown atoms
    """
    [ modification ]
    X1
    [ atoms ]
    A {}
    [ edges ]
    A B
    """,
    # Specify edge between unknown atoms
    """
    [ modification ]
    X2
    [ atoms ]
    B {}
    [ edges ]
    A B
    """,
))
def test_misformed_lines(lines):
    """
    Assure that misformed lines cause the parser to stop.
    """
    lines = textwrap.dedent(lines).splitlines()
    ff = vermouth.forcefield.ForceField(name='test_ff')
    with pytest.raises(IOError):
        vermouth.ffinput.read_ff(lines, ff)



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Add the include statements and the virtual sites for Virtual Site Go model.

The VirtualGoSite model allows to stabilize the ternary structure of Martini
proteins by applying Go potentials maintaining the contacts within the
backbone. The Go potentials are not applied on the backbone beads directly,
instead, they are applied on virtual sites overlapping with the backbone.

The processor defined in this module does not generate the Go potentials.
Instead, they the potential is generated by a third party program. The third
party program generate the interaction matrix for the Go potentials, and the
exclusions as ITP files to be included in the right place in the protein ITP
file. The processor adds an include statement at the end of the `[ exclusions
]` section. Would the third party program need the addition of other include
statements, they can be added by adjusting the `sections` argument of the
processor. To incorporate the include statements, the processor adds the
required lines in the "post_section_lines" meta attribute of the molecules.
This meta attribute is read by :func:`vermouth.gmx.itp.write_molecule_itp`. The
include files are called "<moltype>_<section>_VirtGoSite.itp".

In addition of writing the include statements, the processor adds virtual sites
on top of the backbone beads. The virtual sites are added at the end of the
molecule, they share the residue name, residue id, chain, and position of the
underlying backbone bead. They are also added in the `[ virtual_sitesn ]`
section.
"""

import networkx as nx
from ..molecule import Interaction
from .processor import Processor


class GoVirtIncludes(Processor):
    """
    Add the include statements and the virtual sites for Virtual Site Go model.

    See :mod:`vermouth.processors.go_vs_includes` for more details.

    Every molecule must have a moltype name under the "moltype" key of the
    molecule meta.

    Parameters
    ----------
    sections: collections.abc.Iterable[str], optional
        The sections to which to add an include statement.

    See Also
    --------
    vermouth.processors.name_moltype.NameMolType
        Assign molecule type names to the molecules in a system.
    """
    def __init__(self, sections=('exclusions', )):
        self.sections = sections

    def run_molecule(self, molecule):
        moltype = molecule.meta.get('moltype')
        if not moltype:
            raise ValueError('The molecule does not have a moltype name.')

        add_virtual_sites(molecule, prefix=moltype)

        includes = molecule.meta.get('post_section_lines', {})
        for section in self.sections:
            section_includes = includes.get(section, [])
            section_includes.append('#include "{moltype}_{section}_VirtGoSites.itp"'
                                    .format(moltype=moltype, section=section))
            includes[section] = section_includes
        molecule.meta['post_section_lines'] = includes
        return molecule


def add_virtual_sites(molecule, prefix, backbone='BB', atomname='CA', charge=0):
    """
    Add the virtual sites for GoMartini in the molecule.

    One virtual site is added per backbone bead of the the Martini protein.
    Each virtual site copies the resid, resname, and chain of the backbone
    bead. It also copies the *reference* to the position array, so the virtual
    site position follows if the backbone bead is translated. The virtual sites
    are added *after* all the other atoms of the molecule, each in its own
    charge group, with "CA" as atomname, and a charge of 0. The atomname and
    charge can be set with the `atomname` and `charge` argument, respectively.

    The bead type of the virtual sites is names "<prefix>_<resid>". Where
    `prefix` is provided as an argument of the function, and is expected to be
    the molecule type name.

    Parameters
    ----------
    molecule: vermouth.molecule.Molecule
        The molecule to augment with virtual sites.
    prefix: str
        The prefix to use for bead type names. Usually the molecule type name.
    backbone: str
        The atomname of the backbone beads.
    atomname: str
        The atomname of the virtual sites.
    charge: float or int
        The charge of the virtual sites.
    """
    # If there are no atoms, then there is nothing to do. We can exit early,
    # avoiding to deal with empty iterators.
    if not molecule.nodes:
        return
    virtual_site_nodes = []
    virtual_sites = []
    new_node_id = max(molecule.nodes)
    charge_groups = nx.get_node_attributes(molecule, 'charge_group').values()
    new_charge_group = max(charge_groups) if charge_groups else 0
    for node_id, atom in molecule.nodes(data=True):
        if atom.get('atomname') == backbone:
            new_node_id += 1
            new_charge_group += 1
            virtual_site_nodes.append((new_node_id, {
                'resid': atom['resid'],
                'resname': atom['resname'],
                'atype': '{}_{}'.format(prefix, atom['resid']),
                'charge_group': new_charge_group,
                'chain': atom['chain'],
                'position': atom['position'],
                'atomname': atomname,
                'charge': charge,
            }))
            virtual_sites.append(Interaction(
                atoms=[new_node_id, node_id],
                parameters=['1'],
                meta={'go_vs': True, 'group': 'Virtual go site'},
            ))
    molecule.add_nodes_from(virtual_site_nodes)
    if 'virtual_sitesn' not in molecule.interactions:
        molecule.interactions['virtual_sitesn'] = []
    molecule.interactions['virtual_sitesn'] += virtual_sites



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Handle the RTP format from Gromacs.
"""

import collections
import itertools
import networkx as nx

from ..molecule import Block, Link, Interaction
from .. import utils

__all__ = ['read_rtp']

# Name of the subsections in RTP files.
# Names starting with a '_' are for internal use.
RTP_SUBSECTIONS = ('atoms', 'bonds', 'angles', 'dihedrals',
                   'impropers', 'cmap', 'exclusions',
                   '_bondedtypes')


_BondedTypes = collections.namedtuple(
    '_BondedTypes',
    'bonds angles dihedrals impropers all_dihedrals nrexcl HH14 remove_dih'
)


class _IterRTPSubsectionLines:
    """
    Iterate over the lines of an RTP file within a subsection.
    """

    def __init__(self, parent):
        self.parent = parent
        self.lines = parent.lines
        self.running = True

    def __next__(self):
        if not self.running:
            raise StopIteration
        line = next(self.lines)
        if line.strip().startswith('['):
            self.parent.buffer.append(line)
            self.running = False
            raise StopIteration
        return line

    def __iter__(self):
        return self

    def flush(self):
        """
        Move the iterator after the last line of the subsection.
        """
        for _ in self:
            pass


class _IterRTPSubsections:
    """
    Iterate over the subsection of a RTP file within a section.

    For each subsections, yields its name and  an iterator over its lines.
    """

    def __init__(self, parent):
        self.parent = parent
        self.lines = parent.lines
        self.buffer = collections.deque()
        self.current_subsection = None
        self.running = True

    def __next__(self):
        if not self.running:
            raise StopIteration
        if self.current_subsection is not None:
            self.current_subsection.flush()
        if self.buffer:
            line = self.buffer.popleft()
        else:
            line = next(self.lines)
        stripped = line.strip()
        if stripped.startswith('['):
            # A section header looks like "[ name ]". It matches the following
            # regexp: r"^\s*\[\s*(?P<name>)\s*\]\s*$". Once stripped, the
            # trailing spaces at the beginning and at the end are removed so
            # the string starts with '[' and ends with ']'. The slicing remove
            # these brackets. The final call to strip remove the potential
            # white characters between the brackets and the section name.
            name = stripped[1:-1].strip()
            if name in RTP_SUBSECTIONS:
                subsection = _IterRTPSubsectionLines(self)
                self.current_subsection = subsection
                return name, subsection
            self.parent.buffer.append(line)
            self.running = False
            raise StopIteration
        raise IOError('I am almost sure I should not be here...')

    def __iter__(self):
        return self

    def flush(self):
        """
        Move the iterator after the last subsection of the section.
        """
        for _ in self:
            pass


class _IterRTPSections:
    """
    Iterate over the sections of a RTP file.

    For each section, yields the name of the sections and an iterator over its
    subsections.
    """

    def __init__(self, lines):
        self.lines = lines
        self.buffer = collections.deque()
        self.current_section = None

    def __next__(self):
        if self.current_section is not None:
            self.current_section.flush()
        if self.buffer:
            line = self.buffer.popleft()
        else:
            line = next(self.lines)
        stripped = line.strip()
        if stripped.startswith('['):
            name = stripped[1:-1].strip()
            section = _IterRTPSubsections(self)
            self.current_section = section
            # The "[ bondedtypes ]" is special in the sense that it does
            # not have subsection, but instead have its content directly
            # under it. This breaks the neat hierarchy the rest of the file
            # has. Here we restore the hierarchy by faking that the file
            # contains:
            #
            #    [ bondedtypes ]
            #     [ _bondedtypes ]
            #
            # For now, I do that on the basis of the section name. If other
            # sections are special that way, I'll detect them with a look
            # ahead.
            if name == 'bondedtypes':
                section.buffer.append(' [ _bondedtypes ]')
            return name, section
        raise IOError('Hum... There is a bug in the RTP reader.')

    def __iter__(self):
        return self


def _atoms(subsection, block):
    for line in subsection:
        name, atype, charge, charge_group = line.split()
        atom = {
            'atomname': name,
            'atype': atype,
            'charge': float(charge),
            'charge_group': int(charge_group),
        }
        block.add_atom(atom)


def _base_rtp_parser(interaction_name, natoms):
    def wrapped(subsection, block):
        """
        Parse the lines from a RTP subsection and populate the block.
        """
        interactions = []
        for line in subsection:
            splitted = line.strip().split()
            atoms = splitted[:natoms]
            parameters = splitted[natoms:]
            interactions.append(Interaction(atoms=atoms,
                                            parameters=parameters,
                                            meta={}))
        block.interactions[interaction_name] = interactions
    return wrapped


def _parse_bondedtypes(section):
    # Default taken from
    # 'src/gromacs/gmxpreprocess/resall.cpp::read_resall' in the Gromacs
    # source code.
    defaults = _BondedTypes(bonds=1, angles=1, dihedrals=1,
                            impropers=1, all_dihedrals=0,
                            nrexcl=3, HH14=1, remove_dih=1)

    # The 'bondedtypes' section contains its line directly under it. In
    # order to match the hierarchy model of the rest of the file, the
    # iterator actually yields a subsection named '_bondedtypes'. We need
    # to read the fist line of that first virtual subsection.
    _, lines = next(section)
    line = next(lines)
    read = [int(x) for x in line.split()]

    # Fill with the defaults. The file gives the values in order so we
    # need to append the missing values from the default at the end.
    bondedtypes = _BondedTypes(*(read + list(defaults[len(read):])))

    # Make sure there is no unexpected lines in the section.
    # Come on Jonathan! There must be a more compact way of doing it.
    try:
        next(lines)
    except StopIteration:
        pass
    else:
        raise IOError('"[ bondedtypes ]" section is missformated.')
    try:
        next(section)
    except StopIteration:
        pass
    else:
        raise IOError('"[ bondedtypes ]" section is missformated.')
    return bondedtypes


def _count_hydrogens(names):
    return len([name for name in names if utils.first_alpha(name) == 'H'])


def _keep_dihedral(center, block, bondedtypes):
    if (not bondedtypes.all_dihedrals) and block.has_dihedral_around(center):
        return False
    if bondedtypes.remove_dih and block.has_improper_around(center):
        return False
    return True


def _complete_block(block, bondedtypes):
    """
    Add information from the bondedtypes section to a block.

    Generate implicit dihedral angles, and add function types to the
    interactions.
    """
    block.make_edges_from_interactions()

    # Generate missing dihedrals
    # As pdb2gmx generates all the possible dihedral angles by default,
    # RTP files are written assuming they will be generated. A RTP file
    # have some control over these dihedral angles through the bondedtypes
    # section.
    all_dihedrals = []
    for center, dihedrals in itertools.groupby(
            sorted(block.guess_dihedrals(), key=_dihedral_sorted_center),
            _dihedral_sorted_center):
        if _keep_dihedral(center, block, bondedtypes):
            # TODO: Also sort the dihedrals by index.
            # See src/gromacs/gmxpreprocess/gen_add.cpp::dcomp in the
            # Gromacs source code (see version 2016.3 for instance).
            atoms = sorted(dihedrals, key=_count_hydrogens)[0]
            all_dihedrals.append(Interaction(atoms=atoms, parameters=[], meta={}))
    # TODO: Sort the dihedrals by index
    block.interactions['dihedrals'] = (
        block.interactions.get('dihedrals', []) + all_dihedrals
    )

    # TODO: generate 1-4 interactions between pairs of hydrogen atoms

    # Add function types to the interaction parameters. This is done as a
    # post processing step to cluster as much interaction specific code
    # into this method.
    # I am not sure the function type can be set explicitly in the RTP
    # file except through the bondedtypes section. If it is possible, then the
    # following code can break and atoms can have the function type written
    # twice. Yet, none of the RTP files distributed with Gromacs 2016.3 causes
    # issue.
    functypes = {
        'bonds': bondedtypes.bonds,
        'angles': bondedtypes.angles,
        'dihedrals': bondedtypes.dihedrals,
        'impropers': bondedtypes.impropers,
        'exclusions': 1,
        'cmap': 1,
    }
    for name, interactions in block.interactions.items():
        for interaction in interactions:
            interaction.parameters.insert(0, functypes[name])

    # Set the nrexcl to the block.
    block.nrexcl = bondedtypes.nrexcl


def _split_blocks_and_links(pre_blocks):
    """
    Split all the pre-blocks from `pre_block` into blocks and links.

    Parameters
    ----------
    pre_blocks: dict
        A dict with residue names as keys and instances of :class:`Block`
        as values.

    Returns
    -------
    blocks: dict
        A dict like `pre_block` with all inter-residues information
        stripped out.
    links: list
        A list of instances of :class:`Link` containing the inter-residues
        information from `pre_blocks`.

    See Also
    --------
    _split_block_and_link
        Split an individual pre-block into a block and a link.
    """
    blocks = {}
    links = []
    for name, pre_block in pre_blocks.items():
        block, link = _split_block_and_link(pre_block)
        blocks[name] = block
        links.append(link)
    return blocks, links


def _split_block_and_link(pre_block):
    """
    Split `pre_block` into a block and a link.

    A pre-block is a block as read from an RTP file. It contains both
    intra-residue and inter-residues information. This method split this
    information so that the intra-residue interactions are put in a block
    and the inter-residues ones are put in a link.

    Parameters
    ----------
    pre_block: Block
        The block to split.

    Returns
    -------
    block: Block
        All the intra-residue information.
    link: Link
        All the inter-residues information.
    """
    block = Block(force_field=pre_block.force_field)
    link = Link()

    # It is easier to fill the interactions using a defaultdict,
    # yet defaultdicts are more annoying when reading as querying them
    # creates the keys. So the interactions are revert back to regular
    # dict at the end.
    block.interactions = collections.defaultdict(list)
    link.interactions = collections.defaultdict(list)

    block.name = pre_block.name
    try:
        block.nrexcl = pre_block.nrexcl
    except AttributeError:
        pass

    # Filter the particles from neighboring residues out of the block.
    for atom in pre_block.atoms:
        if not atom['atomname'].startswith('+-'):
            atom['resname'] = pre_block.name
            block.add_atom(atom)
        link.add_node(atom['atomname'])

    # Create the edges of the link and block based on the edges in the pre-block.
    # This will create too many edges in the link, but the useless ones will be
    # pruned latter.
    link.add_edges_from(pre_block.edges)
    block.add_edges_from(edge for edge in pre_block.edges
                         if not any(node[0] in '+-' for node in edge))

    # Split the interactions from the pre-block between the block (for
    # intra-residue interactions) and the link (for inter-residues ones).
    # The "relevant_atoms" set keeps track of what particles are
    # involved in the link. This will allow to prune the link without
    # iterating again through its interactions.
    relevant_atoms = set()
    for name, interactions in pre_block.interactions.items():
        for interaction in interactions:
            for_link = any(atom[0] in '+-' for atom in interaction.atoms)
            if for_link:
                link.interactions[name].append(interaction)
                relevant_atoms.update(interaction.atoms)
            else:
                block.interactions[name].append(interaction)

    # Prune the link to keep only the edges and particles that are
    # relevant.
    nodes = set(link.nodes())
    link.remove_nodes_from(nodes - relevant_atoms)
    # Some interactions do not generate nodes (impropers for instance). If a
    # node is only described in one of such interactions, then the node never
    # appears in the link. Here we make sure these nodes exists even if they
    # are note connected.
    link.add_nodes_from(relevant_atoms)

    # Atoms from a links are matched against a molecule based on its node
    # attributes. The name is a primary criterion, but other criteria can be
    # the residue name (`resname` key) or the residue order in the chain
    # (`order` key). The residue order is 0 when refering to the current
    # residue, +int to refer to residues after the current one in the sequence,
    # -int to refer to a previous residue in the sequence, and '*' for any
    # residue but the current one.
    # RTP files convey the order by prefixing the names with + or -. We need to
    # get rid of these prefixes.
    order = {'+': +1, '-': -1}
    relabel_mapping = {}
    for idx, node in enumerate(link.nodes()):
        atomname = node
        if node[0] in '+-':
            link.nodes[node]['order'] = order[node[0]]
            atomname = atomname[1:]
        else:
            link.nodes[node]['order'] = 0
            link.nodes[node]['resname'] = block.name
        link.nodes[node]['atomname'] = atomname
        relabel_mapping[node] = idx
    nx.relabel_nodes(link, relabel_mapping, copy=False)

    # By relabelling the nodes, we lost the relations between interactions and
    # nodes, so we need to relabel the atoms in the interactions
    new_interactions = collections.defaultdict(list)
    for name, interactions in link.interactions.items():
        for interaction in interactions:
            atoms = tuple(relabel_mapping[atom] for atom in interaction.atoms)
            new_interactions[name].append(Interaction(
                atoms=atoms,
                parameters=interaction.parameters,
                meta=interaction.meta
            ))
    link.interactions = new_interactions


    # Revert the interactions back to regular dicts to avoid creating
    # keys when querying them.
    block.interactions = dict(block.interactions)
    link.interactions = dict(link.interactions)

    return block, link


def _clean_lines(lines):
    # TODO: merge continuation lines
    for line in lines:
        splitted = line.split(';', 1)
        if splitted[0].strip():
            yield splitted[0]


def _dihedral_sorted_center(atoms):
    #return sorted(atoms[1:-1])
    return atoms[1:-1]


def read_rtp(lines, force_field):
    """
    Read blocks and links from a Gromacs RTP file to populate a force field

    Parameters
    ----------
    lines: collections.abc.Iterator
        An iterator over the lines of a RTP file (e.g. a file handle, or a
        list of string).
    force_field: vermouth.forcefield.ForceField
        The force field to populate in place.

    Raises
    ------
    IOError
        Something in the file could not be parsed.
    """
    _subsection_parsers = {
        'atoms': _atoms,
        'bonds': _base_rtp_parser('bonds', natoms=2),
        'angles': _base_rtp_parser('angles', natoms=3),
        'impropers': _base_rtp_parser('impropers', natoms=4),
        'cmap': _base_rtp_parser('cmap', natoms=5),
    }
    # An RTP file contains both the blocks and the links.
    # We first read everything in "pre-blocks"; then we separate the
    # blocks from the links.
    pre_blocks = {}
    bondedtypes = None
    cleaned = _clean_lines(lines)
    for section_name, section in _IterRTPSections(cleaned):
        if section_name == 'bondedtypes':
            bondedtypes = _parse_bondedtypes(section)
            continue
        block = Block(force_field=force_field)
        pre_blocks[section_name] = block
        block.name = section_name

        for subsection_name, subsection in section:
            if subsection_name in _subsection_parsers:
                _subsection_parsers[subsection_name](subsection, block)

    # Pre-blocks only contain the interactions that are explicitly
    # written in the file. Some are incomplete (missing implicit defaults)
    # or must be built from the "bondedtypes" rules.
    for pre_block in pre_blocks.values():
        _complete_block(pre_block, bondedtypes)

    # At this point, the pre-blocks contain both the intra- and
    # inter-residues information. We need to split the pre-blocks into
    # blocks and links.
    blocks, links = _split_blocks_and_links(pre_blocks)

    force_field.blocks.update(blocks)
    force_field.links.extend(links)



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains the Mapping object and the associated parser.
"""
from collections import defaultdict
from functools import partial

from .file_writer import open
from .ffinput import _tokenize, _parse_atom_attributes
from .graph_utils import MappingGraphMatcher
from .log_helpers import StyleAdapter, get_logger
from .molecule import Block
from .parser_utils import SectionLineParser


LOGGER = StyleAdapter(get_logger(__name__))


class Mapping:
    """
    A mapping object that describes a mapping from one resolution to another.

    Attributes
    ----------
    blocks_from: networkx.Graph
        The graph which this :class:`Mapping` object can transform.
    blocks_to: vermouth.molecule.Block
        The :class:`vermouth.molecule.Block` we can transform to.
    references: collections.abc.Mapping
        A mapping of node keys in :attr:`blocks_to` to node keys in
        :attr:`blocks_from` that describes which node in blocks_from should be
        taken as a reference when determining node attributes for nodes in
        blocks_to.
    ff_from: vermouth.forcefield.ForceField
        The forcefield of :attr:`blocks_from`.
    ff_to: vermouth.forcefield.ForceField
        The forcefield of :attr:`blocks_to`.
    names: tuple[str]
        The names of the mapped blocks.
    mapping: dict[collections.abc.Hashable, dict[collections.abc.Hashable, float]]
        The actual mapping that describes for every node key in
        :attr:`blocks_from` to what node key in :attr:`blocks_to` it
        contributes to with what weight.
        ``{node_from: {node_to: weight, ...}, ...}``.

    Note
    ----
    Only nodes described in :attr:`mapping` will be used.

    Parameters
    ----------
    block_from: networkx.Graph
        As per :attr:`blocks_from`.
    block_to: vermouth.molecule.Block
        As per :attr:`blocks_to`.
    mapping: dict[collections.abc.Hashable, dict[collections.abc.Hashable, float]]
        As per :attr:`mapping`.
    references: collections.abc.Mapping
        As per :attr:`references`.
    ff_from: vermouth.forcefield.ForceField
        As per :attr:`ff_from`.
    ff_to: vermouth.forcefield.ForceField
        As per :attr:`ff_to`.
    extra: tuple
        Extra information to be attached to :attr:`blocks_to`.
    normalize_weights: bool
        Whether the weights should be normalized such that the sum of the
        weights of nodes mapping to something is 1.
    names: tuple
        As per :attr:`names`.
    """
    def __init__(self, block_from, block_to, mapping, references,
                 ff_from=None, ff_to=None, extra=(), normalize_weights=False,
                 type='block', names=tuple()):
        self.block_from = block_from.copy()
        self.block_to = block_to.copy()
        self.block_to.extra = extra
        self.block_from.name = names
        self.block_to.name = names
        self.references = references
        self.ff_from = ff_from
        self.ff_to = ff_to
        self.names = names
        self.mapping = mapping
        self.type = type
        # Remove nodes not mapped from blocks_from
        unmapped = set(self.block_from.nodes.keys()) - set(self.mapping.keys())
        self.block_from.remove_nodes_from(unmapped)

        # Normalize the weights
        if normalize_weights:
            self._normalize_weights()

    @property
    def reverse_mapping(self):
        """
        The reverse of :attr:`mapping`.
        ``{node_to: {node_from: weight, ...}, ...}``
        """
        rev_mapping = defaultdict(dict)  # {idx to: {idx from: weight}}
        for idx_from in self.mapping:
            for idx_to, weight in self.mapping[idx_from].items():
                rev_mapping[idx_to][idx_from] = weight
        return dict(rev_mapping)

    def map(self, graph, node_match=None, edge_match=None):
        """
        Performs the partial mapping described by this object on `graph`. It
        first find the induced subgraph isomorphisms between `graph` and
        :attr:`blocks_from`, after which it will process the found isomorphisms
        according to :attr:`mapping`.

        None of the yielded dictionaries will refer to node keys of
        :attr:`blocks_from`. Instead, those will be translated to node keys of
        `graph` based on the found isomorphisms.

        Note
        ----
        Only nodes described in :attr:`mapping` will be used in the
        isomorphism.

        Parameters
        ----------
        graph: networkx.Graph
            The graph on which this partial mapping should be applied.
        node_match: collections.abc.Callable or None
            A function that should take two dictionaries with node attributes,
            and return `True` if those nodes should be considered equal, and
            `False` otherwise. If None, all nodes will be considered equal.
        edge_match: collections.abc.Callable or None
            A function that should take six arguments: two graphs, and four
            node keys. The first two node keys will be in the first graph and
            share an edge; and the last two node keys will be in the second
            graph and share an edge. Should return `True` if a pair of edges
            should be considered equal, and `False` otherwise. If None, all
            edges will be considered equal.

        Yields
        ------
        dict[collections.abc.Hashable, dict[collections.abc.Hashable, float]]
            the correspondence between nodes in `graph` and nodes in
            :attr:`blocks_to`, with the associated weights.
        vermouth.molecule.Block
            :attr:`blocks_to`.
        dict
            :attr:`references` on which :attr:`mapping` has been applied.
        """
        if edge_match is not None:
            edge_match = partial(edge_match, graph, self.block_from)

        return self._graph_map(graph, node_match, edge_match)

    def _graph_map(self, graph, node_match, edge_match):
        """
        Performs the partial mapping described in :meth:`map` using a
        "classical" subgraph isomorphism algorithm.
        """
        # 1 Find subgraph isomorphism between blocks_from and graph
        # 2 Translate found match ({graph idx: blocks from idx}) to indices in
        #   blocks_to using self.mapping
        # 3 Return found matches and blocks_to?
        # PS. I don't really like this, because this object is becoming too
        # intelligent by also having to do the isomorphism. On the other hand,
        # it makes sense from the maths point of view.
        matcher = MappingGraphMatcher(graph, self.block_from,
                                      node_match=node_match,
                                      edge_match=edge_match)
        for match in matcher.subgraph_isomorphisms_iter():
            rev_match = {v: k for k, v in match.items()}

            new_match = defaultdict(dict)
            for graph_idx, from_idx in match.items():
                new_match[graph_idx].update(self.mapping[from_idx])
            references = {out_idx: rev_match[ref_idx]
                          for out_idx, ref_idx in self.references.items()}
            yield dict(new_match), self.block_to, references

    def _normalize_weights(self):
        """
        Normalize weights in :attr:`mapping` such that the sum of the weights
        of nodes mapping to something is 1.
        """
        rev_mapping = self.reverse_mapping
        for idx_from in self.mapping:
            for idx_to, weight in self.mapping[idx_from].items():
                self.mapping[idx_from][idx_to] = weight/sum(rev_mapping[idx_to].values())


class MappingBuilder:
    """
    An object that is in charge of building the arguments needed to create a
    :class:`Mapping` object. It's attributes describe the information
    accumulated so far.

    Attributes
    ----------
    mapping: collections.defaultdict
    blocks_from: None or vermouth.molecule.Block
    blocks_to: None or vermouth.molecule.Block
    ff_from: None or vermouth.forcefield.ForceField
    ff_to: None or vermouth.forcefield.ForceField
    names: list
    references: dict
    """
    def __init__(self):
        self.reset()

    def reset(self):
        """
        Reset the object to a clean initial state.
        """
        self.mapping = defaultdict(dict)
        self.blocks_from = Block()
        self.blocks_to = Block()
        self.ff_from = None
        self.ff_to = None
        self.names = []
        self.references = {}

    def to_ff(self, ff_name):
        """
        Sets :attr:`ff_to`

        Parameters
        ----------
        ff_name
        """
        self.ff_to = ff_name

    def from_ff(self, ff_name):
        """
        Sets :attr:`ff_from`

        Parameters
        ----------
        ff_name
        """
        self.ff_from = ff_name

    @staticmethod
    def _add_block(current_block, new_block):
        """
        Helper method that adds `new_block` to `current_block`, if the latter
        is not `None`. Otherwise, create a new block from `new_block`.

        Parameters
        ----------
        current_block: None or vermouth.molecule.Block
        new_block: vermouth.molecule.Block

        Returns
        -------
        vermouth.molecule.Block
            The combination of `current_block` and `new_block`
        """
        if not current_block:
            current_block = new_block.to_molecule(default_attributes={})
        else:
            current_block.merge_molecule(new_block)
        return current_block

    def add_block_from(self, block):
        """
        Add a block to :attr:`blocks_from`. In addition, apply any 'replace'
        operation described by nodes on themselves::

            {'atomname': 'C', 'charge': 0, 'replace': {'charge': -1}}

        becomes::

            {'atomname': 'C', 'charge': -1}

        Parameters
        ----------
        block: vermouth.molecule.Block
            The block to add.
        """
        block = block.copy()
        for node in block.nodes.values():
            if 'replace' in node:
                node.update(node['replace'])
                del node['replace']
        self.blocks_from = self._add_block(self.blocks_from, block)

    def add_name(self, name):
        """
        Add a name to the mapping.

        Parameters
        ----------
        name: str
            The name to add
        """
        self.names.append(name)

    def add_block_to(self, block):
        """
        Add a block to :attr:`blocks_to`.

        Parameters
        ----------
        block: vermouth.molecule.Block
            The block to add.
        """
        self.blocks_to = self._add_block(self.blocks_to, block)

    def add_node_from(self, attrs):
        """
        Add a single node to :attr:`blocks_from`.

        Parameters
        ----------
        attrs: dict[str]
            The attributes the new node should have.
        """
        if not self.blocks_from:
            self.blocks_from = Block()
            idx = 0
        else:
            idx = max(self.blocks_from.nodes) + 1
        self.blocks_from.add_node(idx, **attrs)

    def add_node_to(self, attrs):
        """
        Add a single node to :attr:`blocks_to`.

        Parameters
        ----------
        attrs: dict[str]
            The attributes the new node should have.
        """
        if not self.blocks_to:
            self.blocks_to = Block()
            idx = 0
        else:
            idx = max(self.blocks_to.nodes) + 1
        self.blocks_to.add_node(idx, **attrs)

    def add_edge_from(self, attrs1, attrs2, edge_attrs):
        """
        Add a single edge to :attr:`blocks_from` between two nodes in
        :attr:`blocks_from` described by `attrs1` and `attrs2`. The nodes
        described should not be the same.

        Parameters
        ----------
        attrs1: dict[str]
            The attributes that uniquely describe a node in
            :attr:`blocks_from`
        attrs2: dict[str]
            The attributes that uniquely describe a node in
            :attr:`blocks_from`
        edge_attrs: dict[str]
            The attributes that should be assigned to the new edge.
        """
        nodes1 = list(self.blocks_from.find_atoms(**attrs1))
        nodes2 = list(self.blocks_from.find_atoms(**attrs2))
        assert len(nodes1) == len(nodes2) == 1
        assert nodes1 != nodes2
        self.blocks_from.add_edge(nodes1[0], nodes2[0], **edge_attrs)

    def add_edge_to(self, attrs1, attrs2, edge_attrs):
        """
        Add a single edge to :attr:`blocks_to` between two nodes in
        :attr:`blocks_to` described by `attrs1` and `attrs2`. The nodes
        described should not be the same.

        Parameters
        ----------
        attrs1: dict[str]
            The attributes that uniquely describe a node in
            :attr:`blocks_to`
        attrs2: dict[str]
            The attributes that uniquely describe a node in
            :attr:`blocks_to`
        edge_attrs: dict[str]
            The attributes that should be assigned to the new edge.
        """
        nodes1 = list(self.blocks_to.find_atoms(**attrs1))
        nodes2 = list(self.blocks_to.find_atoms(**attrs2))
        assert len(nodes1) == len(nodes2) == 1
        assert nodes1 != nodes2
        self.blocks_to.add_edge(nodes1[0], nodes2[0], **edge_attrs)

    def add_mapping(self, attrs_from, attrs_to, weight):
        """
        Add part of a mapping to :attr:`mapping`. `attrs_from` uniquely
        describes a node in :attr:`blocks_from` and `attrs_to` a node in
        :attr:`blocks_to`. Adds a mapping between those nodes with the given
        `weight`.

        Parameters
        ----------
        attrs_from: dict[str]
            The attributes that uniquely describe a node in
            :attr:`blocks_from`
        attrs_to: dict[str]
            The attributes that uniquely describe a node in
            :attr:`blocks_to`
        weight: float
            The weight associated with this partial mapping.
        """
        nodes_from = list(self.blocks_from.find_atoms(**attrs_from))
        nodes_to = list(self.blocks_to.find_atoms(**attrs_to))
        assert len(nodes_from) == len(nodes_to) == 1
        self.mapping[nodes_from[0]][nodes_to[0]] = weight

    def add_reference(self, attrs_to, attrs_from):
        """
        Add a reference to :attr:`references`.

        Parameters
        ----------
        attrs_to: dict[str]
            The attributes that uniquely describe a node in
            :attr:`blocks_to`
        attrs_from: dict[str]
            The attributes that uniquely describe a node in
            :attr:`blocks_from`
        """
        nodes_to = list(self.blocks_to.find_atoms(**attrs_to))
        assert len(nodes_to) == 1
        node_to = nodes_to[0]
        nodes_from = set(self.blocks_from.find_atoms(**attrs_from))
        mapped_nodes = {from_ for from_ in self.mapping if node_to in self.mapping[from_]}
        nodes_from = nodes_from.intersection(mapped_nodes)
        assert len(nodes_from) == 1
        self.references[node_to] = next(iter(nodes_from))

    def get_mapping(self, type):
        """
        Instantiate a :class:`Mapping` object with the information accumulated
        so far, and return it.

        Returns
        -------
        Mapping
            The mapping object made from the accumulated information.
        """
        #if self.blocks_from is None:
        #    return None
        mapping = Mapping(self.blocks_from, self.blocks_to, dict(self.mapping),
                          self.references, ff_from=self.ff_from, ff_to=self.ff_to,
                          type=type, names=tuple(self.names))
        return mapping


class MappingDirector(SectionLineParser):
    """
    A director in charge of parsing the new mapping format. It constructs a new
    :class:`Mapping` object by calling methods of it's builder (default
    :class:`MappingBuilder`) with the correct arguments.

    Parameters
    ----------
    force_fields: dict[str, ForceField]
        Dict of known force fields.
    builder: MappingBuilder

    Attributes
    ----------
    builder
        The builder used to build the :class:`Mapping` object. By default
        :class:`MappingBuilder`.
    identifiers: dict[str, dict[str]]
        All known identifiers at this point. The key is the actual identifier,
        prefixed with either "to\\_" or "from\\_", and the values are the
        associated node attributes.
    section: str
        The name of the section currently being processed.
    from_ff: str
        The name of the forcefield from which this mapping describes a
        transfomation.
    to_ff: str
        The name of the forcefield to which this mapping describes a
        transfomation.
    macros: dict[str, str]
        A dictionary of known macros.
    """
    RESNAME_NUM_SEP = '#'
    """
    The character that separates a resname from a resnumber in shorthand block
    formats.
    """
    RESIDUE_ATOM_SEP = ':'
    """The character that separates a residue identifier from an atomname."""
    COMMENT_CHAR = ';'
    """The character that starts a comment."""
    NO_FETCH_BLOCK = '!'
    """The character that specifies no block should be fetched automatically."""
    SECTION_ENDS = ['block', 'modification']

    def __init__(self, force_fields, builder=None):
        if builder is None:
            self.builder = MappingBuilder()
        else:
            self.builder = builder
        self.force_fields = force_fields
        super().__init__()
        self._reset_mapping()

    def _reset_mapping(self):
        """
        Reinitialize attributes for a new mapping.
        """
        self._current_id = {'from': None, 'to': None}
        self.ff = {'to': None, 'from': None}
        self.identifiers = {}
        self.builder.reset()

    def finalize_section(self, previous_section, ended_section):
        """
        Wraps up parsing of a single mapping.

        Parameters
        ----------
        previous_section: collections.abc.Sequence[str]
            The previously parsed section.
        ended_section: collections.abc.Iterable[str]
            The just finished sections.

        Returns
        -------
        Mapping or None
            The accumulated mapping if the mapping is complete, None otherwise.
        """
        if any(ended in self.SECTION_ENDS for ended in ended_section):
            map_type = previous_section[0]
            mapping = self.builder.get_mapping(map_type)
            self._reset_mapping()
            return mapping
        return None

    def _parse_blocks(self, line):
        """
        Helper method for parsing to_blocks and from_blocks. It parses a line
        containing either a single longhand block description, or multiple
        shorthand block descriptions.

        Parameters
        ----------
        line: str

        yields
        ------
        tuple[str, dict[str]]
            A tuple of an identifier, and it's associated attributes.
        """
        tokens = list(_tokenize(line))
        resid = 0
        if len(tokens) == 2 and tokens[1].startswith('{') and tokens[1].endswith('}'):
            # It's definitely full spec.
            identifier = tokens[0]
            attrs = _parse_atom_attributes(tokens[1])
            yield identifier, attrs
        else:  # It must be shorthand
            for identifier in tokens:
                resname, found_resid = self._parse_block_shorthand(identifier)
                if found_resid is None:
                    resid += 1
                else:
                    resid = found_resid
                if resname.startswith(self.NO_FETCH_BLOCK):
                    resname = resname[len(self.NO_FETCH_BLOCK):]
                attrs = {'resname': resname, 'resid': resid}
                yield identifier, attrs

    def _parse_block_shorthand(self, token):
        """
        Helper method for parsing block shorthand.

        Parameters
        ----------
        token: str

        Returns
        -------
        tuple[str, int]
            A tuple of a resname and a resid.
        """
        if self.RESNAME_NUM_SEP in token:
            # PO4#3 or ALA#2
            resname, resid = token.split(self.RESNAME_NUM_SEP)
            resid = int(resid)
        else:
            # ALA
            resname = token
            resid = None
        return resname, resid

    def _resolve_atom_spec(self, atom_str, prefix=None):
        """
        Helper method that, given an atom token and a prefix ("to" or "from")
        will find the associated node attributes for that atom token. It will
        either separate the given identifier and atomname and look up the
        associated node attributes in :attr:`identifiers`; or take the last
        specified identifier.

        For example, if given the `atom_str` "ALA:CA" it will look up the
        node attributes associated with the identifier "ALA" as specified in
        the (from|to) blocks section, and add them to the attribute
        `{"atomname": "CA"}`. If given the `atom_str` "CA", it will either take
        the previously used identifier if available. If not, it will check
        whether there is only one identifier defined, and use that.

        Parameters
        ----------
        atom_str: str
        prefix: str

        Returns
        -------
        dict[str]
            The node attributes that describe this atom.
        """
        if self.RESIDUE_ATOM_SEP in atom_str:
            id_, name = atom_str.split(self.RESIDUE_ATOM_SEP)
        else:
            id_, name = None, atom_str

        if id_ is None:
            options = {name[1] for name in self.identifiers if name[0] == prefix}
            if len(options) == 1:
                id_ = next(iter(options))

        if id_ is None:
            attrs = self._current_id[prefix].copy()
        else:
            attrs = self.identifiers[(prefix, id_)].copy()
            self._current_id[prefix] = self.identifiers[(prefix, id_)]

        attrs['atomname'] = name

        return attrs

    @SectionLineParser.section_parser('modification', 'to', direction='to')
    @SectionLineParser.section_parser('modification', 'from', direction='from')
    @SectionLineParser.section_parser('block', 'to', direction='to')
    @SectionLineParser.section_parser('block', 'from', direction='from')
    def _ff(self, line, lineno=0, direction=None):
        """
        Parses a "to" or "from" section and sets :attr:`ff`.
        Every line should contain a separate forcefield name.

        Parameters
        ----------
        line: str
        """
        builder_methods = {'to': self.builder.to_ff,
                           'from': self.builder.from_ff}
        self.ff[direction] = line
        builder_methods[direction](line)

    @SectionLineParser.section_parser('modification', 'from blocks',
                                      direction='from', map_type='modification')
    @SectionLineParser.section_parser('modification', 'to blocks',
                                      direction='to', map_type='modification')
    @SectionLineParser.section_parser('block', 'from blocks',
                                      direction='from', map_type='block')
    @SectionLineParser.section_parser('block', 'to blocks',
                                      direction='to', map_type='block')
    def _blocks(self, line, lineno=0, direction=None, map_type=None):
        """
        Parses a "from blocks" or "to_blocks" section and add to
        :attr:`identifiers`. Calls :method:`builder.add_block_from` and
        :method:`builder.add_block_from`.

        A blocks section can be in two formats: shorthand and longhand:
            Shorthand:
                resname_ids := <resname_id>[ <resname_ids>]
                resname_id  := <resname>[#<resid>]
                identifier  == resname_id
                resname     := string that does not contain "#"
                resid       := [0-9]+ (residue index)
            Longhand:
                full_res_specs := <full_res_spec>[\n<full_res_specs>]
                full_res_spec  := [!]<identifier> <res_attrs>
                identifier     := string that does not start with '!'
                res_attrs      := json dict
                An "!" signifies that no block should be fetched based on the
                the resname automatically.

        Identifier *must* be unique for this mapping/direction.

        Examples
        --------
        Shorthand:
            ALA#1 ALA#2
            GLY
        Longhand:
            ALA1 {"resname": "ALA", "resid": 1}
            ALA2 {"resname": "ALA", "resid": 2}

        Parameters
        ----------
        line: str
        lineno: int
        direction: str
            Should be 'to' or 'from'
        map_type: str
            The type of mapping.
        """
        builder_methods = {'to': self.builder.add_block_to,
                           'from': self.builder.add_block_from}
        for identifier, attrs in self._parse_blocks(line):
            fetch = True
            if identifier.startswith(self.NO_FETCH_BLOCK):
                identifier = identifier[len(self.NO_FETCH_BLOCK):]
                fetch = False
            if fetch and attrs.get('resname') is not None:
                block = getattr(self.force_fields[self.ff[direction]], map_type+'s')[attrs['resname']]
                if map_type == 'modification':
                    # Add the modifications to the mapping block_from, so they
                    # participate in the matching criterion.

                    # Caveat emptor
                    ###############
                    # Copying the modification here before modifying it would be
                    # a *good* idea, but that doesn't work (probably because
                    # Links don't have a sane copy method). Better than this
                    # would be to add the modifications attribute at the ff
                    # parser level. This hack-around (and adding it to the ff
                    # parser) probably have unintended side effects.
                    # block = block.copy()
                    for idx in block.nodes:
                        block.nodes[idx]['modifications'] = [block]
                builder_methods[direction](block)
            if direction == 'from':
                if 'resname' in attrs and isinstance(attrs['resname'], str):
                    name = attrs.get('resname')
                else:
                    name = identifier
                self.builder.add_name(name)
            if map_type == 'modification' and 'resname' in attrs:
                del attrs['resname']
            self.identifiers[(direction, identifier)] = attrs

    @SectionLineParser.section_parser('modification', 'from nodes', direction='from')
    @SectionLineParser.section_parser('modification', 'to nodes', direction='to')
    @SectionLineParser.section_parser('block', 'from nodes', direction='from')
    @SectionLineParser.section_parser('block', 'to nodes', direction='to')
    def _nodes(self, line, lineno=0, direction=None):

        """
        Parses a "from nodes" or "to nodes" section. Calls
        :method:`builder.add_node_from` and :method:`builder.add_node_to`.

        atoms      := <atom>[\n<atoms>]
        atom       := [<identifier>:]<atomname>[ <atom_attrs>]
        atomname   := string that does not contain ':'
        atom_attrs := json dict

        If no identifier is specified, the previous one used is taken. If none
        was used before and there is only one option, that is taken. Will add a
        new node with as attributes the union of the attributes associated with
        the identifier and the atom attributes specified. Atom attributes take
        precedence.

        Examples
        --------
        AA1:N {"resname": "ALA"}
        HN
        AA2:CA
        HA {"resid": 2}

        Parameters
        ----------
        line: str
        lineno: int
        direction: str
            Should be 'to' or 'from'
        """
        builder_methods = {'to': self.builder.add_node_to,
                           'from': self.builder.add_node_from}
        name, *new_attrs = _tokenize(line)
        attrs = self._resolve_atom_spec(name, direction)
        if new_attrs:
            new_attrs = _parse_atom_attributes(*new_attrs)
        else:
            new_attrs = {}
        attrs.update(new_attrs)
        builder_methods[direction](attrs)

    @SectionLineParser.section_parser('modification', 'from edges', direction='from')
    @SectionLineParser.section_parser('modification', 'to edges', direction='to')
    @SectionLineParser.section_parser('block', 'from edges', direction='from')
    @SectionLineParser.section_parser('block', 'to edges', direction='to')
    def _edges(self, line, lineno=0, direction=None):
        """
        Parses a "from edges" or "to edges" section. Calls
        :method:`builder.add_edge_from` and :method:`builder.add_edge_to`.

        edges      := <edge>[\n<edges>]
        edge       := <atom> +<atom>[ <edge_attrs>]
        atom       := [<identifier>:]<atomname>
        edge_attrs := json dict

        The combination of attributes defined by the identifier + the specified
        atom name *must* resolve to a unique node in the block.

        Example
        -------
        ALA#1:C ALA#2:N

        Parameters
        ----------
        line: str
        lineno: int
        direction: str
            Should be 'to' or 'from'
        """
        builder_methods = {'from': self.builder.add_edge_from,
                           'to': self.builder.add_edge_to}
        at1, at2, *attrs = _tokenize(line)
        attrs1 = self._resolve_atom_spec(at1, direction)
        attrs2 = self._resolve_atom_spec(at2, direction)
        if attrs:
            attrs = _parse_atom_attributes(*attrs)
        else:
            attrs = {}
        builder_methods[direction](attrs1, attrs2, attrs)

    @SectionLineParser.section_parser('modification', 'mapping')
    @SectionLineParser.section_parser('block', 'mapping')
    def _mapping(self, line, lineno=0):
        """
        Parses a "mapping" section. Calls :method:`builder.add_mapping`.

        mappings := <mapping>[\n<mappings>]
        weight   := float | int
        mapping  := <atom from> <atom to> <weight>

        <atom from> and <atom to> have the same definition as <atom> in the
        "edges" section. See :meth:`_edges`.

        Examples
        --------
        ALA:C    ALA:BB
        H        BB  0

        Parameters
        ----------
        line: str
        lineno: int
        """

        from_, to_, *weight = line.split()
        if weight:
            weight = int(weight[0])
        else:
            weight = 1

        attrs_from = self._resolve_atom_spec(from_, 'from')
        attrs_to = self._resolve_atom_spec(to_, 'to')

        self.builder.add_mapping(attrs_from, attrs_to, weight)

    @SectionLineParser.section_parser('modification', 'reference atoms')
    @SectionLineParser.section_parser('block', 'reference atoms')
    def _reference_atoms(self, line, lineno=0):
        """
        Parses a "reference atom" section. Calls
        :method:`builder.add_reference`.

        reference := <atom to> <atom from>

        Parameters
        ----------
        line: str
        lineno: int
        """
        to_, from_ = line.split()
        attrs_to = self._resolve_atom_spec(to_, 'to')
        attrs_from = self._resolve_atom_spec(from_, 'from')
        self.builder.add_reference(attrs_to, attrs_from)

    @SectionLineParser.section_parser('molecule')
    def _molecule(self, line, lineno=0):
        raise IOError("It look like you're trying to parse an old style "
                      "backmapping file. These some minor modifications to be "
                      "used compared to the files used by `backwards.py`.")


def parse_mapping_file(filepath, force_fields):
    """
    Parses a mapping file.

    Parameters
    ----------
    filepath: str
        The path of the file to parse.
    force_fields: dict[str, ForceField]
        Dict of known forcefields

    Returns
    -------
    list[Mapping]
        A list of all mappings described in the file.
    """
    with open(filepath) as map_in:
        director = MappingDirector(force_fields)
        mappings = list(director.parse(map_in))
    return mappings



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Runs more elaborate integration tests
"""

from collections import OrderedDict
from pathlib import Path
import shlex
import subprocess
import sys
import numpy as np

import pytest
import vermouth
from vermouth.forcefield import ForceField

from .. import datafiles
from ..helper_functions import find_in_path


INTEGRATION_DATA = Path(datafiles.TEST_DATA/'integration_tests')

PATTERN = '{path}/{tier}/{protein}/martinize2/'


MARTINIZE2 = find_in_path()


def assert_equal_blocks(block1, block2):
    """
    Asserts that two blocks are equal to gain the pytest rich comparisons,
    which is lost when doing `assert block1 == block2`
    """
    assert block1.name == block2.name
    assert block1.nrexcl == block2.nrexcl
    assert block1.force_field == block2.force_field  # Set to be equal
    # Assert the order to be equal as well...
    # assert list(block1.nodes) == list(block2.nodes)
    # ... as the attributes
    nodes2 = OrderedDict(block2.nodes(data=True))
    for n_idx, attrs in nodes2.items():
        for k, v in attrs.items():
            if isinstance(v, np.ndarray):
                nodes2[n_idx][k] = pytest.approx(v, abs=1e-3)
    assert OrderedDict(block1.nodes(data=True)) == nodes2
    edges1 = {frozenset(e[:2]): e[2] for e in block1.edges(data=True)}
    edges2 = {frozenset(e[:2]): e[2] for e in block2.edges(data=True)}
    for e, attrs in edges2.items():
        for k, v in attrs.items():
            if isinstance(v, float):
                attrs[k] = pytest.approx(v, abs=1e-3) # PDB precision is 1e-3

    assert edges1 == edges2
    for inter_type, interactions in block1.interactions.items():
        block2_interactions = block2.interactions.get(inter_type, [])
        assert sorted(interactions, key=lambda i: i.atoms) == sorted(block2_interactions, key=lambda i: i.atoms)


def compare_itp(filename1, filename2):
    """
    Asserts that two itps are functionally identical
    """
    dummy_ff = ForceField(name='dummy')
    with open(filename1) as fn1:
        vermouth.gmx.read_itp(fn1, dummy_ff)
    dummy_ff2 = ForceField(name='dummy')
    with open(filename2) as fn2:
        vermouth.gmx.read_itp(fn2, dummy_ff2)
    for block in dummy_ff2.blocks.values():
        block._force_field = dummy_ff
    assert set(dummy_ff.blocks.keys()) == set(dummy_ff2.blocks.keys())
    for name in dummy_ff.blocks:
        block1 = dummy_ff.blocks[name]
        block2 = dummy_ff2.blocks[name]
        assert_equal_blocks(block1, block2)


def compare_pdb(filename1, filename2):
    """
    Asserts that two pdbs are functionally identical
    """
    pdb1 = vermouth.pdb.read_pdb(filename1)
    pdb2 = vermouth.pdb.read_pdb(filename2)
    assert len(pdb1) == len(pdb2)
    for mol1, mol2 in zip(pdb1, pdb2):
        for mol in (mol1, mol2):
            for n_idx in mol:
                node = mol.nodes[n_idx]
                if 'position' in node and node['atomname'] in ('SCN', 'SCP'):
                    # Charge dummies get placed randomly, which complicated
                    # comparisons to no end.
                    # These will be caught by the distances in the edges instead.
                    del node['position']

        assert_equal_blocks(mol1, mol2)


COMPARERS = {'.itp': compare_itp,
             '.pdb': compare_pdb}


def _interaction_equal(interaction1, interaction2):
    """
    Returns True if interaction1 == interaction2, ignoring rounding errors in
    interaction parameters.
    """
    p1 = list(map(float, interaction1.parameters))
    p2 = list(map(float, interaction2.parameters))
    return interaction1.atoms == interaction2.atoms \
           and interaction1.meta == interaction2.meta \
           and pytest.approx(p1) == p2


@pytest.mark.parametrize("tier, protein", [
    ['tier-0', 'mini-protein1_betasheet'],
    ['tier-0', 'mini-protein2_helix'],
    ['tier-0', 'mini-protein3_trp-cage'],
    ['tier-1', 'bpti'],
    ['tier-1', 'lysozyme'],
    ['tier-1', 'villin'],
    # ['tier-2', 'barnase_barstar'],
    # ['tier-2', 'dna'],
    # ['tier-2', 'gpa_dimer'],
])
def test_integration_protein(tmp_path, monkeypatch, tier, protein):
    """
    Runs integration tests by executing the contents of the file `command` in
    the folder tier/protein, and tests whether the contents of the produced
    files are the same as the reference files. The comparison of the files is
    governed by `COMPARERS`.

    Parameters
    ----------
    tmp_path
    tier: str
    protein: str
    """
    monkeypatch.chdir(tmp_path)

    data_path = Path(PATTERN.format(path=INTEGRATION_DATA, tier=tier, protein=protein))

    with open(str(data_path / 'command')) as cmd_file:
        command = cmd_file.read().strip()
    assert command  # Defensive
    command = shlex.split(command)
    result = [sys.executable]
    for token in command:
        if token.startswith('martinize2'):  # Could be martinize2.py
            result.append(str(MARTINIZE2))
        elif token.startswith('.'):
            result.append(str(data_path / token))
        else:
            result.append(token)
    command = result

    proc = subprocess.run(command, cwd='.', timeout=60, check=False,
                          stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE,
                          universal_newlines=True)
    exit_code = proc.returncode
    if exit_code:
        print(proc.stdout)
        print(proc.stderr)
        assert not exit_code

    files = list(tmp_path.iterdir())

    assert files
    assert list(tmp_path.glob('*.itp')), files
    assert list(tmp_path.glob('*.pdb')), files

    for new_file in tmp_path.iterdir():
        filename = new_file.name
        reference_file = data_path/filename
        assert reference_file.is_file()
        ext = new_file.suffix.lower()
        if ext in COMPARERS:
            with monkeypatch.context() as m:
                # Compare Interactions such that rounding errors in the
                # parameters are OK.
                m.setattr(vermouth.molecule.Interaction, '__eq__', _interaction_equal)
                COMPARERS[ext](str(reference_file), str(new_file))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Created on Wed Aug 30 13:07:14 2017

@author: peterkroon
"""

from vermouth import *
import networkx as nx
import numpy as np
import scipy.sparse.csgraph as ssc
import scipy.sparse as ss
import scipy.sparse.linalg as ssl
import scipy.linalg as sl
import scipy.spatial.distance as ssd
import random

import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D


PATH = '../molecules/glkfk.pdb'
mol = read_pdb(PATH)
#mol = read_pdb('../molecules/octane.pdb')


def make_laplacian(adjacency_matrix):
#    return ssc.laplacian(adjacency_matrix, normed=True)
    A = ss.csr_matrix(adjacency_matrix, dtype=np.float64)
    D = ss.diags(np.sum(adjacency_matrix, axis=0), dtype=np.float64, format='csc')
    D_inv_sqrt = np.sqrt(ssl.inv(D))
    # Using an unweighted Laplacian does RatioCut (equal degree/size)
    # using a weighted Laplacian does NCut (equal volume, maximize intracluster similarity)
    L = D - A
    # D^(-1/2) * L * D^(-1/2); but then matrix multiplication
    # This works with the * operator, because they're sparse matrices.
    L = D_inv_sqrt * L * D_inv_sqrt
    assert np.allclose(L.A, L.T.A), 'Laplacian is not symmetric'
    return L


def spectral(G, ndim=3):
    # https://networkx.github.io/documentation/networkx-2.0/reference/generated/networkx.drawing.layout.spectral_layout.html
#    conmat = nx.adjacency_matrix(G, weight='weight')
#    L = make_laplacian(conmat.A)
#    print(L.A)
    L = nx.normalized_laplacian_matrix(G, weight='weight')
    vals, vecs = ssl.eigsh(L, k=ndim+1, sigma=-1e-7, which='LM')
#    S = np.linalg.eig(L.A)
#    order = np.argsort(S[0])
#    S = np.real(S[0]), np.real(S[1])
#    vals = S[0][order]
#    vecs = S[1][:, order]
    scale = 1/np.sqrt(vals[1:ndim+1])
    return vecs[:, 1:ndim+1] * scale

def MDS(G, fill=10, ndim=3):
    # https://en.wikipedia.org/wiki/Multidimensional_scaling
    n = len(G)
    distmat = nx.adjacency_matrix(G, weight='distance').A
    distmat[distmat == 0] = fill
    D2 = distmat ** 2
    J = np.eye(n) - 1/n * np.ones(n) * np.ones(n).T
    # Must be matmul
    B = - 0.5 * J @ D2 @ J
    S = np.linalg.eig(B)
    order = np.argsort(S[0])[::-1]
    vals = S[0][order]
    vecs = S[1][:, order]
    M = np.real(vecs[:, :ndim])
    L = np.diag(np.sqrt(vals[:ndim]))
    return M @ L

def smacof_dense(G, eps=1e-1, maxiter=5000, ndim=3, start=None):
    # https://en.wikipedia.org/wiki/Stress_majorization
    n = len(G)
    if start is None:
        X = np.random.rand(n, ndim)
    else:
        X = start
    disparities = nx.adjacency_matrix(G, weight='distance').A
    w = np.zeros_like(disparities)
    w[disparities != 0] = 1
    w *= (n*(n-1))/(w * disparities**2).sum()
    V = np.diag(np.sum(w, axis=1))
    V -= w
    V = np.linalg.pinv(V)
    for it in range(maxiter):
        dis = ssd.squareform(ssd.pdist(X))
        stress = np.sum(w * (dis - disparities)**2)
        if stress < eps:
            print('Converged after {} iterations'.format(it))
            break
        dis[dis == 0] = np.inf
        # Update X using the Guttman transform
        S = 1/dis
        B = np.diag(np.sum(w * disparities * S, axis=1))
        B -= w * disparities * S
        displacement = V @ B
        X = displacement @ X

    return X


def smacof(G, eps=1e-1, maxiter=5000, ndim=3, start=None):
    # https://en.wikipedia.org/wiki/Stress_majorization
    n = len(G)
    if start is None:
        X = np.random.rand(n, ndim)
    else:
        X = start
    disparities = nx.adjacency_matrix(G, weight='distance')
    data = [n*(n-1)/disparities.power(2).sum()] * disparities.nnz
    w = ss.csc_matrix((data, disparities.nonzero()), shape=disparities.shape)

    # The pseudo-inverse will be dense anyway. And let's make sure it's an array.
    V = np.diag(w.sum(axis=0).A[0])
    V -= w
    V = np.linalg.pinv(V).A
    for it in range(maxiter):
        vecs = (X[w.nonzero()[0]] - X[w.nonzero()[1]])
        # vecs is dense. I don't think this can be avoided.
        # dis however, is sparse.
        dis = np.sqrt(np.sum(vecs**2, axis=-1))
        dis = ss.csr_matrix((dis, w.nonzero()), w.shape)
        stress = (w * (dis - disparities).power(2)).sum()
        if stress < eps:
            print('Converged after {} iterations'.format(it))
            break
        S = dis.power(-1)
        # Update X using the Guttman transform
        B = -w.multiply(disparities).multiply(S)
        diag = np.array(-B.sum(axis=0))[0]
        B.setdiag(diag)
        # V is dense, so displacements is dense.
        displacement = V @ B
        X = displacement @ X

    return X

def dist_to_weight(distance):
    return np.exp(-distance**2)

#def draw(mat):
#    fig = plt.figure()
#    ax = fig.add_subplot(111, projection='3d')
#    ax.scatter(mat[:, 0], mat[:, 1], mat[:, 2])

for idx, jdx in mol.edges():
    d = mol[idx][jdx].get('distance', 1)
    mol[idx][jdx]['distance'] = d
    mol[idx][jdx]['weight'] = dist_to_weight(d)

paths = []
for target in nx.all_pairs_shortest_path(mol, cutoff=2):
    for path in target[1].values():
        if len(path) == 3:
            paths.append(path)

mol_angle = mol.copy()
angle = np.deg2rad(180-109)
for idx, jdx, kdx in paths:
    l1 = mol[idx][jdx]['distance']
    l2 = mol[jdx][kdx]['distance']
    length = l1**2 * np.sin(angle)**2 + (l2 + l1*np.cos(angle))**2
    length = np.sqrt(length)
    mol_angle.add_edge(idx, kdx, distance=length, weight=dist_to_weight(length))
print('Added {} angles.'.format(len(paths)))

#embedded = MDS(mol_angle, fill=1000)
embedded = spectral(mol_angle)
embedded = smacof_dense(mol_angle, maxiter=5000, start=embedded)
#embedded = smacof_dense(mol_angle, eps=2e-2, maxiter=5000)
#embedded = smacof(mol_angle, maxiter=5000, start=embedded)

orig_pos = np.array([mol.node[idx]['position'] for idx in mol.node])

for idx in mol.node:
    mol.node[idx]['position'] = embedded[idx]
draw(mol, node_size=30, with_label=True)
n = 5

edges = [random.choice(list(mol.edges())) for _ in range(n)]
print(edges)

for idx, jdx in edges:
#    d1 = np.linalg.norm(mol.node[idx]['position'] - mol.node[jdx]['position'])
    d1 = np.linalg.norm(orig_pos[idx] - orig_pos[jdx])
    d2 = np.linalg.norm(embedded[idx] - embedded[jdx])
    print(idx, jdx, d1, d2)



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unittests for the PDB writer.
"""

# Pylint is wrongly complaining about fixtures.
# pylint: disable=redefined-outer-name


import numpy as np
import networkx as nx

import pytest

import vermouth
from vermouth.molecule import Molecule
from vermouth.pdb import pdb


@pytest.fixture
def dummy_system():
    """
    Build a system with 3 dummy molecules.

    All nodes have the following attributes:

    * `atomname`
    * `resname`
    * `resid`
    * `chain` set to `''`
    * `element` set to the value of `atomname`
    * `positions`

    Some nodes have a `charge` attribute set to an integer charge. The
    molecules are connected.
    """
    nodes = (
        {'atomname': 'A', 'resname': 'A', 'resid': 1, },
        {'atomname': 'B', 'resname': 'A', 'resid': 1, 'charge': -1},
        {'atomname': 'C', 'resname': 'A', 'resid': 2, 'charge': 1},
        {'atomname': 'D', 'resname': 'A', 'resid': 3, },
        {'atomname': 'E', 'resname': 'A', 'resid': 4, },
        {'atomname': 'F', 'resname': 'B', 'resid': 4, },
        {'atomname': 'G', 'resname': 'B', 'resid': 4, },
        {'atomname': 'H', 'resname': 'B', 'resid': 4, },
    )
    edges = [(0, 1), (2, 3), (4, 5), (5, 6), (5, 7)]
    graph = nx.Graph()
    for idx, node in enumerate(nodes):
        node['chain'] = ''
        node['element'] = node['atomname']
        node['position'] = np.array([1, 2, -3])
        graph.add_node(idx, **node)
    graph.add_edges_from(edges)
    molecules = [Molecule(graph.subgraph(component))
                 for component in nx.connected_components(graph)]
    system = vermouth.system.System()
    system.molecules = molecules
    return system


@pytest.fixture
def missing_pos_system(dummy_system):
    """
    Build a system of dummy molecules with some nodes lacking a `position`
    attribute.
    """
    for node in dummy_system.molecules[1].nodes.values():
        del node['position']
    return dummy_system


def test_conect(dummy_system):
    """
    Test that the CONECT record is written as expected. Test also that the
    overall format is correct.
    """
    pdb_found = pdb.write_pdb_string(dummy_system, conect=True, omit_charges=False)
    expected = '''
ATOM      1 A    A       1      10.000  20.000 -30.000  1.00  0.00          A   
ATOM      2 B    A       1      10.000  20.000 -30.000  1.00  0.00          B 1-
TER       3      A       1 
ATOM      4 C    A       2      10.000  20.000 -30.000  1.00  0.00          C 1+
ATOM      5 D    A       3      10.000  20.000 -30.000  1.00  0.00          D   
TER       6      A       3 
ATOM      7 E    A       4      10.000  20.000 -30.000  1.00  0.00          E   
ATOM      8 F    B       4      10.000  20.000 -30.000  1.00  0.00          F   
ATOM      9 G    B       4      10.000  20.000 -30.000  1.00  0.00          G   
ATOM     10 H    B       4      10.000  20.000 -30.000  1.00  0.00          H   
TER      11      B       4 
CONECT    1    2
CONECT    4    5
CONECT    7    8
CONECT    8    9   10
END
'''
    assert pdb_found.strip() == expected.strip()


def test_write_failure_missing_pos(missing_pos_system):
    """
    Make sure the writing fails when coordinates are missing and
    `nan_missing_pos` is not set. (Shall be `False` by default.)
    """
    with pytest.raises(KeyError):
        pdb.write_pdb_string(missing_pos_system)


def test_write_success_missing_pos(missing_pos_system):
    """
    Make sure the writing succeed when coordinates are missing and
    `nan_missing_pos` is `True`.
    """
    pdb_found = pdb.write_pdb_string(
        missing_pos_system,
        conect=False,
        omit_charges=False,
        nan_missing_pos=True,  # Argument of interest!
    )
    expected = '''
ATOM      1 A    A       1      10.000  20.000 -30.000  1.00  0.00          A   
ATOM      2 B    A       1      10.000  20.000 -30.000  1.00  0.00          B 1-
TER       3      A       1 
ATOM      4 C    A       2         nan     nan     nan  1.00  0.00          C 1+
ATOM      5 D    A       3         nan     nan     nan  1.00  0.00          D   
TER       6      A       3 
ATOM      7 E    A       4      10.000  20.000 -30.000  1.00  0.00          E   
ATOM      8 F    B       4      10.000  20.000 -30.000  1.00  0.00          F   
ATOM      9 G    B       4      10.000  20.000 -30.000  1.00  0.00          G   
ATOM     10 H    B       4      10.000  20.000 -30.000  1.00  0.00          H   
TER      11      B       4 
END
'''
    assert pdb_found.strip() == expected.strip()



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from vermouth.parser_utils import SectionParser

# pylint: disable=missing-function-docstring, missing-class-docstring, no-member, invalid-name

def test_metaclass_inheritance():
    """
    Test that the (populated) METH_DICT from the SectionParser metaclass is
    correctly inherited or reset by subclasses
    """
    class A(metaclass=SectionParser):
        def method_a(self):
            pass

        @SectionParser.section_parser('section_a')
        def method_a2(self):
            pass

    class B(A):
        @SectionParser.section_parser('section_b')
        def method_b1(self):
            pass

    class C(A):
        @SectionParser.section_parser('section_c')
        def method_c1(self):
            pass

    assert A.METH_DICT == {('section_a', ): (A.method_a2, {})}
    assert B.METH_DICT == {('section_a',): (A.method_a2, {}), ('section_b',): (B.method_b1, {})}
    assert C.METH_DICT == {('section_a',): (A.method_a2, {}), ('section_c',): (C.method_c1, {})}

    class C2(C):
        @SectionParser.section_parser('section_c')
        def method_c2(self):
            pass

    assert C.METH_DICT == {('section_a',): (A.method_a2, {}), ('section_c',): (C.method_c1, {})}
    assert C2.METH_DICT == {('section_a',): (A.method_a2, {}), ('section_c',): (C2.method_c2, {})}



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the :class:`vermouth.processors.go_vs_includes.GoVirtIncludes` processor.
"""

# Pylint issues false warnings because of pytest's fixtures.
# pylint: disable=redefined-outer-name

import pytest
from vermouth import Molecule
from vermouth.processors import GoVirtIncludes

@pytest.fixture(params=(0, 4))
def molecule_for_go(request):
    """
    Molecule to test virtual site-based GoMartini.

    The molecule has a variable number of residue set in the fixture parameter.
    Each residue has a backbone and a side chain beads. It also has some post
    sections lines that should not disapear when running the processor, and a
    moltype name as this is required.
    """
    n_residues = request.param
    molecule = Molecule()
    for residue_idx in range(n_residues):
        base_atom_idx = residue_idx * 2
        for relative_atom_idx, atomname in enumerate(('BB', 'SC1')):
            molecule.add_node(
                base_atom_idx + relative_atom_idx,
                resid=residue_idx,
                resname='XX',
                atomname=atomname,
                charge_group=base_atom_idx,
                chain='A',
                # Should be an array, but it is not relevant for the test
                position=[0, 0, 0],
            )
    molecule.meta['post_section_lines'] = {
        'exclusions': ['I exist'],
        'other': ['Hello'],
    }
    molecule.meta['moltype'] = 'TEST'
    return molecule


def test_go_virt_includes(molecule_for_go):
    """
    Test the processor in normal conditions.
    """
    molecule = molecule_for_go
    initial_n_backbone = len(molecule.nodes) // 2  # Each residues has 2 atoms
    processor = GoVirtIncludes()
    processor.run_molecule(molecule)

    expected_post = {
        'exclusions': ['I exist', '#include "TEST_exclusions_VirtGoSites.itp"'],
        'other': ['Hello'],
    }

    assert molecule.meta['post_section_lines'] == expected_post
    # Each residue in the initial molecule has a backbone and a side chain
    # beads. We create one virtual site per backbone beads, so we expect to
    # have a side chain and a virtual site for each backbone bead.
    assert len(molecule.nodes) == initial_n_backbone * 3


def test_go_virt_includes_no_moltype():
    """
    Test the processor raises an exception if the moltype is not defined.
    """
    molecule = Molecule()
    processor = GoVirtIncludes()
    with pytest.raises(ValueError):
        processor.run_molecule(molecule)



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides a processor that generates positions for every charge dummy.
"""
import itertools
import operator

import numpy as np

from .processor import Processor

DEFAULT_DUMMY_ATTRIBUTE = 'charge_dummy'


def fibonacci_sphere(n_samples):
    """
    Place points near-evenly distributed on a sphere.

    Use the Fibonacci sphere algorithm to place 'n_samples' points at the
    surface of a sphere of radius 1, centered on the origin.

    Parameters
    ----------
    n_samples: int
        Number of points to place.

    Returns
    -------
    numpy.ndarray
        3D coordinates of the points.
    """
    offset = 2 / n_samples
    increment = np.pi * (3 - np.sqrt(5))
    sample_idx = np.arange(n_samples)
    y = (sample_idx * offset - 1) + offset / 2  # pylint: disable=invalid-name
    r = np.sqrt(1 - y * y)  # pylint: disable=invalid-name
    phi = (sample_idx % n_samples) * increment
    x = np.cos(phi) * r  # pylint: disable=invalid-name
    z = np.sin(phi) * r  # pylint: disable=invalid-name
    return np.stack([x, y, z]).T


def colinear_pair():
    """
    Build two points on a line around the origin at a random orientation.
    """
    vector = np.random.rand(3)
    vector /= np.linalg.norm(vector)
    points = np.stack([np.zeros((3, )), vector])
    points -= vector/2
    return points


def find_anchor(molecule, node_key, attribute_tag=DEFAULT_DUMMY_ATTRIBUTE):
    """
    Find the non-dummy bead to which a charge dummy is anchored.

    Each charge dummy has to be attached to exactly one non-dummy atom. This
    function returns the node key for that non-dummy atom.

    Parameters
    ----------
    molecule: networkx.Graph
        The molecule to work on.
    node_key:
        The node key of the charge dummy.
    attribute_tag: str
        The name of the atom attribute used to describe charge dummies.

    Returns
    -------
    collections.abc.Hashable
        The node key of the anchor in the molecule graph.

    Raises
    ------
    ValueError
        Raised if there are no anchor, or more than one anchor, found. Raised
        also if the charge dummy is not a charge dummy.
    """
    dummy = molecule.nodes[node_key]
    if dummy.get(attribute_tag, None) is None:
        msg = 'Node "{}" is not a charge dummy. Check the "{}" node attribute.'
        raise ValueError(msg.format(node_key, attribute_tag))

    # There should be only one anchor, and that anchor is the (hopefully) only
    # neighbor that is not a dummy.
    potential_anchors = [
        neighbor
        for neighbor in molecule.neighbors(node_key)
        if molecule.nodes[neighbor].get(attribute_tag, None) is None
    ]
    if not potential_anchors:
        raise ValueError('No anchor found for dummy bead "{}": {}.'
                         .format(node_key, molecule.nodes[node_key]))
    elif len(potential_anchors) > 1:
        raise ValueError('Too many potential anchors found for dummy "{}" ({} found).'
                         .format(node_key, len(potential_anchors)))

    return potential_anchors[0]


def locate_dummy(molecule, anchor_key, dummy_keys, attribute_tag=DEFAULT_DUMMY_ATTRIBUTE):
    """
    Set the position of a group of charge dummies around a non-dummy anchor.

    The molecule is modified in-place.

    The charge dummies are placed at a distance to the anchor defined in nm by
    their charge dummy attribute, the name of which is given in the
    'attribute_tag' argument.

    Parameters
    ----------
    molecule: vermouth.molecule.Molecule
        The molecule to work on.
    anchor_key:
        The key of the non-dummy anchor all the charge dummies are connected to.
    dummy_keys: collections.abc.Iterable
        A collection of atom keys for charge dummies to position.
    attribute_tag: str
        Name of the atom attribute that describe charge dummies.
    """
    anchor_position = molecule.nodes[anchor_key].get('position')
    if anchor_position is None:
        msg = 'The anchor of the "{}" dummy ("{}") does not have a position.'
        raise ValueError(msg.format(anchor_key, anchor_position[0]))

    distances = []
    distance_error_keys = []
    for dummy_key in dummy_keys:
        try:
            distances.append(float(molecule.nodes[dummy_key].get(attribute_tag)))
        except ValueError:
            distance_error_keys.append(dummy_key)
    if distance_error_keys:
        msg = ('The following charge dummies have an invalid for their {} '
               'attribute: {}. The values have to be numbers.'
               .format(attribute_tag, ', '.join(distance_error_keys)))
        raise ValueError(msg)
    distances = np.array(distances)

    if len(dummy_keys) == 2:
        points = colinear_pair()
    else:
        points = fibonacci_sphere(len(dummy_keys))
    points *= distances[:, None]

    for dummy_key, position in zip(dummy_keys, points):
        molecule.nodes[dummy_key]['position'] = position + anchor_position


def locate_all_dummies(molecule, attribute_tag=DEFAULT_DUMMY_ATTRIBUTE):
    """
    Set the position of all charge dummies of a molecule.

    The molecule is modified in-place.

    The charge dummies are placed at a distance to the anchor defined in nm by
    their charge dummy attribute, the name of which is given in the
    'attribute_tag' argument.

    Parameters
    ----------
    molecule: vermouth.molecule.Molecule
        The molecule to work on.
    attribute_tag: str
        Name of the atom attribute that describe charge dummies.
    """

    dummies = [
        (find_anchor(molecule, dummy_key, attribute_tag), dummy_key)
        for dummy_key in molecule.nodes
        if molecule.nodes[dummy_key].get(attribute_tag, None) is not None
    ]
    dummies.sort()  # Sort primarily on the anchor key as it appears first

    grouped_by_anchor = itertools.groupby(dummies, key=operator.itemgetter(0))
    for anchor_key, dummy_pairs in grouped_by_anchor:
        dummy_keys = [pair[1] for pair in dummy_pairs]
        locate_dummy(molecule, anchor_key, dummy_keys, attribute_tag)

class LocateChargeDummies(Processor):
    def __init__(self, attribute_tag=DEFAULT_DUMMY_ATTRIBUTE):
        super().__init__()
        self.attribute_tag = attribute_tag

    def run_molecule(self, molecule):
        locate_all_dummies(molecule, attribute_tag=self.attribute_tag)
        return molecule



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for the AddMoleculeEdgesAtDistance processor.
"""

# The redefined-outer-name check from pylint wrongly catches the use of pytest
# fixtures.
# pylint: disable=redefined-outer-name

import pytest

import vermouth
from vermouth.pdb.pdb import read_pdb
from vermouth.processors.add_molecule_edges import (
    DNA_ACCEPTORS, DNA_DONORS, DNA_HB_DIST,
)
from .datafiles import SHORT_DNA


@pytest.fixture
def short_dna():
    """
    Build a system that contains a short DNA double strand without edges.
    """
    molecules = read_pdb(SHORT_DNA)

    system = vermouth.system.System()
    system.molecules = molecules
    return system


def short_dna_general(short_dna):
    """
    DNA double strands with hydrogen bonds added using
    :class:`vermouth.AddMoleculeEdgesAtDistance`.
    """
    processor = vermouth.AddMoleculeEdgesAtDistance(
        DNA_HB_DIST, DNA_DONORS, DNA_ACCEPTORS
    )
    system = processor.run_system(short_dna)
    return system


def short_dna_strands(short_dna):
    """
    DNA double strands with hydrogen bonds added using
    :class:`vermouth.MergeNucleicStrands`.
    """
    processor = vermouth.MergeNucleicStrands()
    system = processor.run_system(short_dna)
    return system


@pytest.fixture(params=(short_dna_general, short_dna_strands))
def short_dna_edges(request, short_dna):
    """
    Create successively a system with edges produced by
    :class:`vermouth.AddMoleculeEdgesAtDistance` and by
    :class:`vermouth.MergeNucleicStrands`. Having this fixture allows to run
    :func:`test_add_molecule_edges_distance` with the output of both
    :func:`short_dna_general` and :func:`short_dna_strands` the outputs of
    which should be identical.
    """
    return request.param(short_dna)


def test_add_molecule_edges_distance(short_dna_edges):
    """
    Assure that :class:`vermouth.AddMoleculeEdgesAtDistance` and
    :class:`vermouth.MergeNucleicStrands` adds the expected edges.
    """
    expected = set([
        (52, 74),
        (10, 117),
        (54, 73),
        (13, 114),
        (35, 92),
        (32, 95),
        (51, 76),
        (33, 93),
        (11, 115),
        (73, 74),
        (32, 33),
        (52, 54),
        (93, 95),
        (114, 115),
        (11, 13),
    ])
    assert set(short_dna_edges.molecules[0].edges) == expected



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains unittests for vermouth.ismags.
"""

# no-member because module networkx does indeed have a member isomorphism;
# redefined-outer-name because pylint does not like fixtures;
# protected-access because it's tests.
# pylint: disable=no-member, redefined-outer-name, protected-access

from pprint import pprint
from time import perf_counter

from hypothesis import given, note, settings, event
import hypothesis.strategies as st
from hypothesis_networkx import graph_builder

import networkx as nx
import pytest

import vermouth.ismags
from vermouth.graph_utils import categorical_maximum_common_subgraph as MCS

from .helper_functions import make_into_set


def basic_molecule(node_data, edge_data=None):
    """
    Construct a simple Molecule based with specified nodes and edges.
    """
    if edge_data is None:
        edge_data = {}

    mol = vermouth.Molecule()
    for idx, node in enumerate(node_data):
        mol.add_node(idx, **node)
    for (idx, jdx), data in edge_data.items():
        mol.add_edge(idx, jdx, **data)
    return mol


@pytest.fixture(params=[
    (
        [(0, dict(name='a')),
         (1, dict(name='a')),
         (2, dict(name='b')),
         (3, dict(name='b')),
         (4, dict(name='a')),
         (5, dict(name='a'))],
        [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5)]
    ),
    (
        range(1, 5),
        [(1, 2), (2, 4), (4, 3), (3, 1)]
    ),
    (
        [],
        [(0, 1), (1, 2), (2, 3), (3, 4), (4, 5), (5, 0), (0, 6), (6, 7),
         (2, 8), (8, 9), (4, 10), (10, 11)]
    ),
    (
        [],
        [(0, 1), (1, 2), (1, 4), (2, 3), (3, 5), (3, 6)]
    ),
    (
        [],
        [(0, 1), (0, 5), (1, 2), (1, 3), (1, 4), (1, 6), (2, 7)]
    ),
    (
        [],
        nx.petersen_graph().edges,
    ),
])
def graphs(request):
    """
    Some simple, symmetric graphs
    """
    nodes, edges = request.param
    graph = nx.Graph()
    graph.add_nodes_from(nodes)
    graph.add_edges_from(edges)
    return graph


def test_symmetric_self_isomorphism(graphs):
    """
    Make sure that when considering symmetry, there is only one isomorphism
    between a graph and itself
    """
    ismags = vermouth.ismags.ISMAGS(graphs, graphs)
    iso = list(ismags.find_isomorphisms(True))
    assert make_into_set(iso) == make_into_set([{n: n for n in graphs}])

    graph_matcher = nx.isomorphism.GraphMatcher(graphs, graphs)
    nx_answer = list(graph_matcher.isomorphisms_iter())
    assert make_into_set(iso) <= make_into_set(nx_answer)


def test_asymmetric_self_isomorphism(graphs):
    """
    Compare with reference implementation
    """
    ismags = vermouth.ismags.ISMAGS(graphs, graphs)
    ismags_answer = list(ismags.find_isomorphisms(False))
    graph_matcher = nx.isomorphism.GraphMatcher(graphs, graphs)
    nx_answer = list(graph_matcher.isomorphisms_iter())
    assert make_into_set(ismags_answer) == make_into_set(nx_answer)


def test_broken_edgecase():
    """
    In this edgecase the ordering of the nodes matters for the symmetries
    found. This is a bad thing. It happens, because _refine_node_partitions in
    _couple_nodes does *not* switch node orders, causing it to produce an
    invalid coupling, losing out on a permutation.
    """
    graph = nx.Graph()
    nx.add_path(graph, range(5))
    graph.add_edges_from([(2, 5), (5, 6)])

    ismags = vermouth.ismags.ISMAGS(graph, graph)
    ismags_answer = list(ismags.find_isomorphisms(True))
    assert len(ismags_answer) == 1

    graph = nx.relabel_nodes(graph, {0: 0, 1: 1, 2: 2, 3: 3, 4: 6, 5: 4, 6: 5})
    ismags = vermouth.ismags.ISMAGS(graph, graph)
    ismags_answer = list(ismags.find_isomorphisms(True))
    assert len(ismags_answer) == 1

# no-value-for-parameter because `draw` is not explicitely passed;
# no-member because module networkx does indeed have a member isomorphism.
# pylint: disable=no-value-for-parameter, no-member


MAX_NODES = 5
ATTRNAMES = ['attr1', 'attr2']
NODE_DATA = st.dictionaries(keys=st.sampled_from(ATTRNAMES),
                            values=st.integers(min_value=0, max_value=MAX_NODES))

ATTRS = st.lists(st.sampled_from(ATTRNAMES), unique=True, min_size=0, max_size=2)
ISO_DATA = st.dictionaries(keys=st.sampled_from(ATTRNAMES),
                           values=st.integers(max_value=MAX_NODES, min_value=0))

ISO_BUILDER = graph_builder(node_data=ISO_DATA, min_nodes=0, max_nodes=MAX_NODES,
                            edge_data=ISO_DATA,
                            node_keys=st.integers(max_value=MAX_NODES, min_value=0))
MCS_BUILDER = graph_builder(node_data=ISO_DATA, min_nodes=0, max_nodes=5,
                            node_keys=st.integers(max_value=MAX_NODES, min_value=0))


@settings(deadline=500)
@given(subgraph=ISO_BUILDER, attrs=st.one_of(st.none(), ATTRS))
def test_hypo_symmetric_self_isomorphism(subgraph, attrs):
    """
    Make sure that when considering symmetry, there is only one isomorphism
    between a graph and itself
    """
    if attrs is None:
        node_match = lambda n1, n2: True
    else:
        node_match = nx.isomorphism.categorical_node_match(attrs, [None]*len(attrs))

    note(("Graph nodes", subgraph.nodes(data=True)))
    note(("Graph edges", subgraph.edges(data=True)))

    ismags = vermouth.ismags.ISMAGS(subgraph, subgraph, node_match=node_match,
                                    edge_match=node_match)

    found = make_into_set(ismags.find_isomorphisms(True))
    note(("Found", found))

    assert found == make_into_set([{n: n for n in subgraph}])


@settings(deadline=500)
@given(graph=ISO_BUILDER, subgraph=ISO_BUILDER, attrs=st.one_of(st.none(), ATTRS))
def test_isomorphism_nonmatch(graph, subgraph, attrs):
    """
    Test against networkx reference implementation using graphs that are
    probably not subgraphs without considering symmetry.
    """

    if attrs is None:
        node_match = lambda n1, n2: True
    else:
        node_match = nx.isomorphism.categorical_node_match(attrs, [None]*len(attrs))

    note(("Graph nodes", graph.nodes(data=True)))
    note(("Graph edges", graph.edges(data=True)))
    note(("Subgraph nodes", subgraph.nodes(data=True)))
    note(("Subgraph edges", subgraph.edges(data=True)))

    ref_time = perf_counter()
    matcher = nx.isomorphism.GraphMatcher(graph, subgraph, node_match=node_match,
                                          edge_match=node_match)
    expected = make_into_set(matcher.subgraph_isomorphisms_iter())
    ref_time -= perf_counter()


    a_ism_time = perf_counter()
    ismags = vermouth.ismags.ISMAGS(graph, subgraph, node_match=node_match,
                                    edge_match=node_match)
    asymmetric = make_into_set(ismags.find_isomorphisms(False))
    a_ism_time -= perf_counter()
    s_ism_time = perf_counter()
    ismags = vermouth.ismags.ISMAGS(graph, subgraph, node_match=node_match,
                                    edge_match=node_match)
    symmetric = make_into_set(ismags.find_isomorphisms(True))
    s_ism_time -= perf_counter()

    note(("Symmetric", symmetric))
    note(("Asymmetric", asymmetric))
    note(("Expected", expected))

    if a_ism_time < ref_time:
        event('Asymmetric ISMAGS faster than reference')
    if s_ism_time < a_ism_time:
        event('Symmetric ISMAGS faster than asymmetric')
    if s_ism_time < ref_time:
        event('Symmetric ISMAGS faster than reference')

    assert asymmetric == expected
    assert symmetric <= asymmetric
    if symmetric == asymmetric and expected:
        assert ismags.analyze_symmetry(subgraph,
                                       ismags._sgn_partitions,
                                       ismags._sge_colors) == ([], {})
    elif symmetric != asymmetric:
        assert ismags.analyze_symmetry(subgraph, ismags._sgn_partitions,
                                       ismags._sge_colors) != ([], {})


@settings(deadline=500)
@given(st.data())
def test_isomorphism_match(data):
    """
    Test against networkx reference implementation using graphs that are
    subgraphs without considering symmetry.
    """
    attrs = data.draw(st.one_of(st.none(), ATTRS))
    if attrs is None:
        node_match = lambda n1, n2: True
    else:
        node_match = nx.isomorphism.categorical_node_match(attrs, [None]*len(attrs))

    graph = data.draw(ISO_BUILDER)

    if graph:
        nodes = data.draw(st.sets(st.sampled_from(list(graph.nodes)),
                                  max_size=len(graph)))
    else:
        nodes = []
    subgraph = graph.subgraph(nodes)

    note(("Graph nodes", graph.nodes(data=True)))
    note(("Graph edges", graph.edges(data=True)))
    note(("Subgraph nodes", subgraph.nodes(data=True)))
    note(("Subgraph edges", subgraph.edges(data=True)))

    ref_time = perf_counter()
    matcher = nx.isomorphism.GraphMatcher(graph, subgraph, node_match=node_match,
                                          edge_match=node_match)
    expected = make_into_set(matcher.subgraph_isomorphisms_iter())
    ref_time -= perf_counter()

    a_ism_time = perf_counter()
    ismags = vermouth.ismags.ISMAGS(graph, subgraph, node_match=node_match,
                                    edge_match=node_match)
    asymmetric = make_into_set(ismags.find_isomorphisms(False))
    a_ism_time -= perf_counter()
    s_ism_time = perf_counter()
    ismags = vermouth.ismags.ISMAGS(graph, subgraph, node_match=node_match,
                                    edge_match=node_match)
    symmetric = make_into_set(ismags.find_isomorphisms(True))
    s_ism_time -= perf_counter()

    note(("Symmetric", symmetric))
    note(("Asymmetric", asymmetric))
    note(("Expected", expected))

    if a_ism_time < ref_time:
        event('Asymmetric ISMAGS faster than reference')
    if s_ism_time < a_ism_time:
        event('Symmetric ISMAGS faster than asymmetric')
    if s_ism_time < ref_time:
        event('Symmetric ISMAGS faster than reference')

    assert asymmetric == expected
    assert symmetric <= asymmetric
    if symmetric == asymmetric and expected:
        assert ismags.analyze_symmetry(subgraph,
                                       ismags._sgn_partitions,
                                       ismags._sge_colors) == ([], {})
    elif symmetric != asymmetric:
        assert ismags.analyze_symmetry(subgraph, ismags._sgn_partitions,
                                       ismags._sge_colors) != ([], {})


@settings(deadline=500)
@given(graph=MCS_BUILDER, subgraph=MCS_BUILDER, attrs=st.one_of(st.none(), ATTRS))
def test_mcs_nonmatch(graph, subgraph, attrs):
    """
    Test against networkx reference implementation using graphs that are
    probably not subgraphs without considering symmetry.
    """
    if attrs is None:
        node_match = lambda n1, n2: True
        attrs = []
    else:
        node_match = nx.isomorphism.categorical_node_match(attrs, [None]*len(attrs))

    note(("Graph nodes", graph.nodes(data=True)))
    note(("Graph edges", graph.edges(data=True)))
    note(("Subgraph nodes", subgraph.nodes(data=True)))
    note(("Subgraph edges", subgraph.edges(data=True)))

    ref_time = perf_counter()
    expected = make_into_set(MCS(graph, subgraph, attributes=attrs))
    ref_time -= perf_counter()

    a_ism_time = perf_counter()
    ismags = vermouth.ismags.ISMAGS(graph, subgraph, node_match=node_match)
    asymmetric = make_into_set(ismags.largest_common_subgraph(False))
    a_ism_time -= perf_counter()
    s_ism_time = perf_counter()
    ismags = vermouth.ismags.ISMAGS(graph, subgraph, node_match=node_match)
    symmetric = make_into_set(ismags.largest_common_subgraph(True))
    s_ism_time -= perf_counter()

    note(("Symmetric", symmetric))
    note(("Asymmetric", asymmetric))
    note(("Expected", expected))

    if a_ism_time < ref_time:
        event('Asymmetric ISMAGS faster than reference')
    if s_ism_time < a_ism_time:
        event('Symmetric ISMAGS faster than asymmetric')
    if s_ism_time < ref_time:
        event('Symmetric ISMAGS faster than reference')

    assert asymmetric == expected or not expected
    assert symmetric <= asymmetric
#    if symmetric == asymmetric and expected:
#        assert ismags.analyze_symmetry(subgraph,
#                                       ismags._sgn_partitions,
#                                       ismags._sge_colors) == ([], {})


@settings(deadline=500)
@given(st.data())
def test_mcs_match(data):
    """
    Test against networkx reference implementation using graphs that are
    subgraphs without considering symmetry.
    """
    attrs = data.draw(st.one_of(st.none(), ATTRS))
    if attrs is None:
        node_match = lambda n1, n2: True
        attrs = []
    else:
        node_match = nx.isomorphism.categorical_node_match(attrs, [None]*len(attrs))

    graph = data.draw(MCS_BUILDER)

    if graph:
        nodes = data.draw(st.sets(st.sampled_from(list(graph.nodes)),
                                  max_size=len(graph)))
    else:
        nodes = []
    subgraph = graph.subgraph(nodes)

    note(("Graph nodes", graph.nodes(data=True)))
    note(("Graph edges", graph.edges(data=True)))
    note(("Subgraph nodes", subgraph.nodes(data=True)))
    note(("Subgraph edges", subgraph.edges(data=True)))

    ref_time = perf_counter()
    expected = make_into_set(MCS(graph, subgraph, attributes=attrs))
    ref_time -= perf_counter()

    a_ism_time = perf_counter()
    ismags = vermouth.ismags.ISMAGS(graph, subgraph, node_match=node_match)
    asymmetric = make_into_set(ismags.largest_common_subgraph(False))
    a_ism_time -= perf_counter()
    s_ism_time = perf_counter()
    ismags = vermouth.ismags.ISMAGS(graph, subgraph, node_match=node_match)
    symmetric = make_into_set(ismags.largest_common_subgraph(True))
    s_ism_time -= perf_counter()

    note(("Symmetric", symmetric))
    note(("Asymmetric", asymmetric))
    note(("Expected", expected))

    if a_ism_time < ref_time:
        event('Asymmetric ISMAGS faster than reference')
    if s_ism_time < a_ism_time:
        event('Symmetric ISMAGS faster than asymmetric')
    if s_ism_time < ref_time:
        event('Symmetric ISMAGS faster than reference')

    assert asymmetric == expected or not expected
    assert symmetric <= asymmetric
#    if symmetric == asymmetric and expected:
#        assert ismags.analyze_symmetry(subgraph,
#                                       ismags._sgn_partitions,
#                                       ismags._sge_colors) == ([], {})


@pytest.mark.parametrize('node_data1, edges1, node_data2, edges2, attrs, expected', [
    ([], {}, [], {}, ['id'], [{}]),
    (
        [{'id': 1}], {},
        [{'id': 1}], {},
        ['id'],
        [{0: 0}]
    ),
    (
        [{'id': 1}, {'id': 2}], {},
        [{'id': 1}, {'id': 2}], {},
        ['id'],
        [{0: 0, 1: 1}]
    ),
    (
        [{'id': 1}, {'id': 2}], {(0, 1): {}},
        [{'id': 1}, {'id': 2}], {},
        ['id'],
        [{0: 0}, {1: 1}]
    ),
    (
        [{'id': 1}], {},
        [{'id': 1}, {'id': 1}], {(0, 1): {}},
        ['id'],
        [{0: 0}, {0: 1}]
    ),
    (
        [{'id': 1}, {'id': 1}], {(0, 1): {}},
        [{'id': 1}], {},
        ['id'],
        [{0: 0}, {1: 0}]
    ),
    (
        [{'id': 0}, {'id': 1}, {'id': 2}, {'id': 3}, {'id': 2}],
        {(0, 1): {}, (1, 2): {}, (2, 3): {}, (3, 4): {}},
        [{'id': 1}, {'id': 2}, {'id': 3}, {'id': 4}, {'id': 5}],
        {(0, 1): {}, (1, 2): {}, (2, 3): {}, (1, 4): {}},
        ['id'],
        [{1: 0, 2: 1, 3: 2}]
    ),
    (
        [{}, {}, {}, {}, {}],
        {(0, 1): {}, (1, 2): {}, (2, 3): {}, (2, 4): {}},
        [{}, {}, {}, {}, {}],
        {(0, 1): {}, (1, 2): {}, (2, 3): {}, (3, 4): {}},
        [],
        [{3: 2, 4: 0, 2: 1, 0: 4},
         {0: 0, 1: 1, 4: 3, 2: 2},
         {2: 2, 1: 3, 4: 1, 0: 4},
         {0: 1, 3: 4, 1: 2, 2: 3},
         {3: 0, 0: 3, 2: 1, 1: 2},
         {3: 2, 0: 0, 2: 3, 4: 4},
         {1: 2, 0: 3, 2: 1, 4: 0},
         {2: 2, 1: 3, 3: 1, 0: 4},
         {2: 2, 0: 0, 3: 3, 1: 1},
         {4: 2, 0: 0, 3: 4, 2: 3},
         {0: 1, 4: 4, 2: 3, 1: 2},
         {3: 0, 4: 2, 2: 1, 0: 4}]

    ),
])
def test_maximum_common_subgraph_known_outcome(node_data1, edges1, node_data2, edges2, attrs, expected):
    """
    Tests for the function ``maximum_common_subgraph``.
    """
    graph1 = basic_molecule(node_data1, edges1)
    graph2 = basic_molecule(node_data2, edges2)
    ism = vermouth.ismags.ISMAGS(graph1, graph2, node_match=nx.isomorphism.categorical_node_match(attrs, [None]*len(attrs)))
    found = ism.largest_common_subgraph(False)
    found = make_into_set(found)
    expected = make_into_set(expected)
    print("Number of found answers", len(found))
    print("Number of expected answers", len(expected))

    pprint(("Found answers", found))
    pprint(("Expected answers", expected))

    pprint(("Expected answers that are not found", expected - found))
    assert found == expected


@pytest.mark.parametrize('node_data1, edges1, node_data2, edges2, expected', [
    (
        [{'atomname': 0, 'element': 0}, {'atomname': 0, 'element': 0}],
        {(0, 1): {}},
        [{'atomname': 0, 'element': 0}, {'atomname': 0, 'element': 0}],
        {(0, 1): {}},
        [{0:0, 1:1}]
    ),
    (
        [{'atomname': 0, 'element': 0}, {'atomname': 0, 'element': 0},
         {'atomname': 0, 'element': 0}],
        {(0, 1): {}, (1, 2): {}},
        [{'atomname': 0, 'element': 0}, {'atomname': 0, 'element': 0},
         {'atomname': 0, 'element': 0}],
        {(0, 1): {}, (1, 2): {}},
        [{0: 0, 1: 1, 2: 2}]),
    (
        [{'atomname': 0, 'element': 0}, {'atomname': 0, 'element': 0},
         {'atomname': 0, 'element': 0}],
        {(0, 1): {}, (1, 2): {}},
        [{'atomname': 0, 'element': 0}, {'atomname': 0, 'element': 0}],
        {(0, 1): {}},
        [{0: 0, 1: 1}, {1: 0, 2: 1}]
    ),
    (
        [{'atomname': 0, 'element': 0}, {'atomname': 0, 'element': 0},
         {'atomname': 0, 'element': 0}, {'atomname': 0, 'element': 0}],
        {(0, 1): {}, (1, 2): {}, (1, 3): {}},
        [{'atomname': 0, 'element': 0}, {'atomname': 0, 'element': 0},
         {'atomname': 0, 'element': 0}],
        {(0, 1): {}, (1, 2): {}},
        [{0: 0, 1: 1, 2: 2}, {0: 0, 1: 1, 3: 2}, {1: 1, 2: 0, 3: 2}]
    ),
])
def test_isomorphism_known_outcome(node_data1, edges1, node_data2, edges2, expected):
    """
    Tests for the function ``isomorphism``. May give a false failure (!) if
    a different (but still correct) answer is returned due to an implementation
    change.
    """
    reference = basic_molecule(node_data1, edges1)
    graph = basic_molecule(node_data2, edges2)
    ism = vermouth.ismags.ISMAGS(reference, graph, node_match=nx.isomorphism.categorical_node_match('element', None))
    found = list(ism.subgraph_isomorphisms_iter())
    found = make_into_set(found)
    expected = make_into_set(expected)
    assert found == expected
